/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//


#ifndef _LEVELOPACITY_H_
#define _LEVELOPACITY_H_

#include "CFIVS.H"
#include "LevelData.H"
#include "ProblemDomain.H"


#define STENSIZE D_TERM(1,*2,*2)

struct CFStruct{
  CFIVS d[2*CH_SPACEDIM];
};

class LevelOpacity
{
public:

  LevelOpacity();

  virtual ~LevelOpacity();

  void define(const DisjointBoxLayout& a_grids,
              const ProblemDomain&     a_domain);

  ///
  /**
     a_opacity (tau+/tau-) : output 2 vector: integration from edge 
       of a_grids into the interior, assuming opacity at the edge of 
       the grid level is assumed to be zero (opaque away
       from refined region assumption).  Requires one ghost cell.

     a_density (rho): scalar 
     a_magneticField (B): SPACEDIM vector
     a_dx: grid spacing
  */
  void computeOpacity(LevelData<FArrayBox>&       a_opacity,
                      const LevelData<FArrayBox>& a_density,
                      const LevelData<FArrayBox>& a_magneticField,
                      const RealVect              a_dx) const ;

private:

  virtual void integrateBox(FArrayBox& a_tau, 
                            int a_comp, 
                            const FArrayBox& a_rho, 
                            const FArrayBox& a_B, 
                            const Box& a_region, 
                            const RealVect& a_dx) const ;

  void maxOrder(IntVect& a_order, 
                IntVect& a_sign, 
                IntVect& a_sample,
                const RealVect& a_vector) const;

  void stencil(Real* a_weight, IntVect* a_iv,
               Real& length,   const RealVect&  a_dx,   
               const IntVect&  a_ordering,
               const IntVect&  a_upstream, 
               const RealVect& B,
               const IntVect&  a_loc) const;

 
  LayoutData<CFStruct> m_cfivs;
  ProblemDomain        m_domain;

};

#endif
