
#include <iostream>
#include <fstream>

using namespace std;

#undef CH_USE_HDF5

#include "IntVect.H"
#include "Box.H"
#include "BoxIterator.H"
#include "BoxLayout.H"
#include "FArrayBox.H"
#include "ProblemDomain.H"
#include "CylinderBL.H"
#include "GeometryShop.H"
#include "EBIndexSpace.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"
#include "EBArith.H"

#include "EBAMRPoissonOp.H"
#include "EBAMRHelmholtzOp.H"
#include "EBAMRNSviscOp.H"

#include "NeumannPoissonDomainBC.H"
//#include "NeumannPoissonDomainRZBC.H"
#include "NeumannPoissonEBBC.H"
//#include "NeumannPoissonEBRZBC.H"
#include "NeumannFunction.H"
//#include "NeumannRZFunction.H"

#include "DirichletPoissonDomainBC.H"
//#include "DirichletPoissonDomainRZBC.H"
#include "DirichletPoissonEBBC.H"
//#include "DirichletPoissonEBRZBC.H"
#include "DirichletFunction.H"
//#include "DirichletRZFunction.H"

#include "DirichletHelmholtzDomainBC.H"
#include "DirichletHelmholtzEBBC.H"
#include "DirichletVectFunctionHH.H"
#include "DirichletVectFunctionHHRZ.H"

#include "NeumannHelmholtzDomainBC.H"
#include "NeumannHelmholtzEBBC.H"
#include "NeumannVectFunctionHH.H"
#include "NeumannVectFunctionHHRZ.H"

#include "DirichletNSviscDomainBC.H"
#include "DirichletNSviscEBBC.H"
#include "DirichletVectFunction.H"
#include "DirichletRZVectFunction.H"

#include "BiCGStabSolver.H"
#include "EBSimpleSolver.H"

#include "UnionIF.H"
#include "PlaneIF.H"
#include "IntersectionIF.H"
#include "ImplicitFunctionBL.H"
#include "ComplementIF.H"
#include "LatheIF.H"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

void get_polygons( Vector<Vector<RealVect> >& polygons, const char* name );

double testrzwk( double r,double z, double mu, double lambda );


class Dbgfuns
{
   double eps, epsi, eps2i;
   double a, b, r0;

   double pol7( double x )
   {
      if( x >= 0 && x <= 1 )
         return (x*x*x*x*(0.25-0.6*x+0.5*x*x-x*x*x/7))*140;
      else if( x > 1 )
         return 1;
      else
         return 0;
   }

   double pol7p( double x )
   {
      if( x >= 0 && x <= 1 )
         return 140*x*x*x*(1-x)*(1-x)*(1-x);
      else if( x > 1 )
         return 0;
      else
         return 0;
   }

   double pol7pp( double x )
   {
      if( x >= 0 && x <= 1 )
         return 3*140*x*x*(1-x)*(1-x)*(1-2*x);
      else if( x > 1 )
         return 0;
      else
         return 0;
   }
  
public:
  Dbgfuns( double ein, double lo, double hi )
  {
    eps = ein;
    a = lo;
    b = hi;
    r0 = 0.5;
    epsi = 1/eps;
    eps2i=epsi*epsi;

  }
  double smfun( double x )
  {
     return pol7(x*epsi);
  }
  double smfunp( double x )
  {
     return pol7p(x*epsi)*epsi;
  }
  double smfunpp( double x )
  {
     return pol7pp(x*epsi)*eps2i;
  }
  double smfun2( double x )
  {
     return smfun(x-a)*smfun(b-x);   
  }
  double smfun2p( double x )
  {
     return smfunp(x-a)*smfun(b-x)-smfun(x-a)*smfunp(b-x);   
  }

  double smfun2pp( double x )
  {
     return smfunpp(x-a)*smfun(b-x)-2*smfunp(x-a)*smfunp(b-x)+smfun(x-a)*smfunpp(b-x);   
  }

  double fi( RealVect x )
  {
     double rad = sqrt(x[0]*x[0]+x[1]*x[1]);
     return smfun(rad-r0)*smfun2(x[0])*smfun2(x[1]);
  }

  void fivals( RealVect x, double& fi, double& fix, double& fiy,
	       double& fixx, double& fixy, double& fiyy )
  {

     double rad = sqrt(x[0]*x[0]+x[1]*x[1]);
     double rx  = x[0]/rad;
     double ry  = x[1]/rad;
     double rxx = 1/rad-x[0]*x[0]/(rad*rad*rad);
     double rxy =-x[0]*x[1]/(rad*rad*rad);
     double ryy = 1/rad-x[1]*x[1]/(rad*rad*rad);

     fi  = smfun(rad-r0)*smfun2(x[0])*smfun2(x[1]);

     fix = rx*smfunp(rad-r0)*smfun2(x[0] )*smfun2(x[1])+
                        smfun(rad-r0 )*smfun2p(x[0])*smfun2(x[1]);

     fiy = ry*smfunp(rad-r0)*smfun2(x[0] )*smfun2(x[1])+
                         smfun(rad-r0 )*smfun2(x[0] )*smfun2p(x[1]);

     fixx= rx*rx*smfunpp(rad-r0)*smfun2(x[0])*smfun2(x[1]) +
                                 rxx*smfunp(rad-r0 )*smfun2(x[0])*smfun2(x[1]) +
                                2*rx*smfunp(rad-r0 )*smfun2p(x[0])*smfun2(x[1]) +
                                     smfun(rad-r0  )*smfun2pp(x[0])*smfun2(x[1]);

     fixy= rx*ry*smfunpp(rad-r0)*smfun2(x[0])*smfun2(x[1]) +
                                 rxy*smfunp(rad-r0 )*smfun2(x[0])*smfun2(x[1]) +
		                  rx*smfunp(rad-r0 )*smfun2(x[0])*smfun2p(x[1]) +
		                  ry*smfunp(rad-r0 )*smfun2p(x[0])*smfun2(x[1]) +
		                     smfun(rad-r0  )*smfun2p(x[0])*smfun2p(x[1]);

     fiyy= ry*ry*smfunpp(rad-r0)*smfun2(x[0])*smfun2(x[1]) +
                                 ryy*smfunp(rad-r0 )*smfun2(x[0])*smfun2(x[1]) +
		                2*ry*smfunp(rad-r0 )*smfun2(x[0])*smfun2p(x[1]) +
		                     smfun(rad-r0  )*smfun2(x[0])*smfun2pp(x[1]);
  }
  
};


void  get_testdata( LevelData<EBCellFAB>& u, const EBISLayout& ebis,
		    const RealVect dx, const RealVect x0, int nr )
//-----------------------------------------------------------------------
//
// Test function, nr =1 --> sin(x)*cos(y) (scalar)
//                nr =2 --> (sin(x)*cos(y), cos(x)*sin(y)) vector valued
//                nr =3 --> (fi*sin(x)*cos(y), fi*cos(x)*sin(y)) vector valued,
//                         with fi a cutting function which forces zero on boundary.
//                nr =4 --> (x*sin(x)*cos(y), cos(x)*sin(y)) vector valued,
//                         with (1/x)*(d x*u/dx)=0 for x=0, good test for
//                          cylindrical case.
//                nr = 5 --> same as nr=4, but for the Laplace operator (scalar).
//
//-----------------------------------------------------------------------  
{

   for( DataIterator li=ebis.getDisjointLayout().dataIterator() ; li.ok() ; ++li )
   {
      DataIndex dex(li());
      //      cout << "box in EBISBox " << ebis[dex].getRegion() << endl;
      //      cout << "box in u       " << u[dex].getRegion() << endl;
      int vnr = 1;
      for( BoxIterator b(u[dex].getRegion()); b.ok(); ++b )
      {
	//         if( !ebis[dex].isCovered(b()) )
	 {
	     Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
             for( int vi=0 ; vi < vofs.size() ; vi++ )
  	     {
	        RealVect x=static_cast<RealVect>(b());
                x[0] = (x[0]+0.5)*dx[0]+x0[0];
	        x[1] = (x[1]+0.5)*dx[1]+x0[1];
		if( nr == 1 )
    	           u[dex](vofs[vi],0) = sin(x[0])*cos(x[1]);
                else if( nr == 2 )
		{
    	           u[dex](vofs[vi],0) = sin(x[0])*cos(x[1]);
		   u[dex](vofs[vi],1) = cos(x[0])*sin(x[1]);
		}
		else if( nr == 3 )
		{
                   Dbgfuns fobj( 1.0, -2.0, 2.0 );
                   double fi=fobj.fi(x);
    	           u[dex](vofs[vi],0) = fi*sin(x[0])*cos(x[1]);
		   u[dex](vofs[vi],1) = fi*cos(x[0])*sin(x[1]);
		}
                else if( nr == 4 )
		{
    	           u[dex](vofs[vi],0) = x[0]*sin(x[0])*cos(x[1]);
		   u[dex](vofs[vi],1) = cos(x[0])*sin(x[1]);
		}
                else if( nr == 5 )
    	           u[dex](vofs[vi],0) = x[0]*sin(x[0])*cos(x[1]);
                else if( nr == 6 )
		   u[dex](vofs[vi],0) = vnr++;
                else if( nr == 7 )
		{
		   u[dex](vofs[vi],0) = 0;
                   if( vofs[vi].gridIndex()[0] == 0 && vofs[vi].gridIndex()[1] == 3 )
		      u[dex](vofs[vi],0) = 1;
		}
	     }
	 }
      }
   }
}


void  get_rho( LevelData<EBCellFAB>& rho, const EBISLayout& ebis,
	       const RealVect dx, const RealVect x0, int nr )
//-----------------------------------------------------------------------
//
// Test rho, nr =1 -->  Identically 1
//           nr =2 -->  jump between 78.58 and 1.482e-7 as in NS-solver.
//
//-----------------------------------------------------------------------  
{

   for( DataIterator li=ebis.getDisjointLayout().dataIterator() ; li.ok() ; ++li )
   {
      DataIndex dex(li());
      for( BoxIterator b(rho[dex].getRegion()); b.ok(); ++b )
      {
	 {
	     Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
             for( int vi=0 ; vi < vofs.size() ; vi++ )
  	     {
	        RealVect x=static_cast<RealVect>(b());
                x[0] = (x[0]+0.5)*dx[0]+x0[0];
	        x[1] = (x[1]+0.5)*dx[1]+x0[1];
		if( nr == 1 )
		   rho[dex](vofs[vi],0) = 1;
		else if( nr == 2 )
		{
		   if( x[1] < 0 )
		      rho[dex](vofs[vi],0) = 78.58;
		   else
		      rho[dex](vofs[vi],0) = 1.482e-7;
		}
	     }
	 }
      }
   }
}
//-----------------------------------------------------------------------
void read_pars( char* filnam, int& testcase, int& relax_type, int& bc_order,
		int& ebgeom, int& domainnr, Real& k2, Real& mu, Real& la,
		bool& rz, int& ni, int& nj, int& nimax, int& njmax, Real &cof,
		int& sol, int& nit, int& levels, int& n_level_ghost, int& n_ebis_ghost,
		int& dirichlet, bool& resmulvol )
{
   FILE *finp = fopen( filnam, "r" );
   if( finp == NULL )
   {
     printf("Error opening file '%s' for reading \n", filnam );
     exit(2);
   }
    
   int cylon;
   int cylcut = 0;
   fscanf(finp, "%i  %i  %i %i \n", &testcase, &cylon, &relax_type, &bc_order );
 
   int cnt = 0;
   int safety_limit = 100000;
   int zeroboundary =0;
   int nozz = 0;
   int rmv = 0;
   while( getc(finp) != '\n' && cnt < safety_limit ) {cnt++;};
   fscanf(finp,"%lf %lf %lf %lf \n",&k2, &mu, &la, &cof );
 
   while( getc(finp) != '\n' && cnt < safety_limit ) {cnt++;};
   fscanf(finp,"%i %i %i %i %i\n",&ni, &nj, &nimax, &njmax, &levels );

   while( getc(finp) != '\n' && cnt < safety_limit ) {cnt++;};
   fscanf(finp,"%i %i \n",&n_level_ghost, &n_ebis_ghost );

   while( getc(finp) != '\n' && cnt < safety_limit ) {cnt++;};
   fscanf(finp,"%i %i %i %i \n", &dirichlet, &zeroboundary, &cylcut, &nozz );

   while( getc(finp) != '\n' && cnt < safety_limit ) {cnt++;};
   fscanf(finp,"%i %i %i \n",&sol, &nit, &rmv );
 
   fclose(finp);

   rz =  testcase == 8 || testcase == 9 || testcase == -1 || testcase == 10
     || testcase == 11 || testcase == 13 || testcase == -5 || testcase == -6;
   if( rz && !zeroboundary )
   {
      domainnr = 2;
      ebgeom = 1;
   }
   else if( rz && zeroboundary )
   {
      domainnr = 3;
      ebgeom = 1;
   }
   else
   {
     domainnr = 1;
     ebgeom = 0;
   }

   if( cylcut == 1 )
   {
     ebgeom = 3;
   }
   else if( cylcut == 2 )
   {
     ebgeom = 4;
   }

   if( cylon == 0 )
      ebgeom = 2;
   if( nozz == 1 )
   {
     domainnr = 4;
     ebgeom   = 5;
   }
   if( nozz == 2 )
   {
     domainnr = 5;
     ebgeom   = 5;
   }

   if( levels < 0 || levels > 3 )
   {
     cout << "Error in level parameter " << endl;
     levels = 1;
   }
   resmulvol = rmv == 1;
}


//-----------------------------------------------------------------------
void get_exactres( LevelData<EBCellFAB>& ex, const EBISLayout& ebis,
		   const RealVect dx, const RealVect x0, int nr, int divmul,
		   Real k2, Real mu, Real la )

//-----------------------------------------------------------------------
// Exact residual corresponding to given test data.
//
//-----------------------------------------------------------------------
{
   Dbgfuns fobj( 1.0, -2.0, 2.0 );
   int nc = ex.nComp();

   for( DataIterator li=ebis.getDisjointLayout().dataIterator() ; li.ok() ; ++li )
   {
      DataIndex dex(li());
      for( BoxIterator b(ex[dex].getRegion()); b.ok(); ++b )
	 if( !ebis[dex].isCovered(b()) )
	 {
	    Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
            if( vofs.size() > 1 )
                cout << "voff voff" << endl;

            for( int vi=0 ; vi < vofs.size() ; vi++ )
	    {
	       RealVect x=ebis[dex].centroid(vofs[vi])+static_cast<RealVect>(b());
               x[0] = (x[0]+0.5)*dx[0]+x0[0];
	       x[1] = (x[1]+0.5)*dx[1]+x0[1];

               if( nr == 1 )
                  ex[dex](vofs[vi],0) = -2*mu*sin(x[0])*cos(x[1]);
               else if( nr == 2 )
                  ex[dex](vofs[vi],0) = -(2*mu+k2)*sin(x[0])*cos(x[1]);
               else if( nr == 3 )
                  ex[dex](vofs[vi],0) = -(2*mu+k2)*sin(x[0])*cos(x[1]) +
		    mu*cos(x[0])*cos(x[1])/x[0];
               else if( nr == 4 )
	       {
                  double sc = sin(x[0])*cos(x[1]);
		  //  	          double ss = sin(x[0])*sin(x[1]);
                  double cs = cos(x[0])*sin(x[1]);
		  //		  double cc = cos(x[0])*cos(x[1]);
                  ex[dex](vofs[vi],0) = (-sc)*(2*mu+la) +
		       mu*(-sc)+(la+mu)*(-sc) - sc/k2;
                  ex[dex](vofs[vi],1) = (mu+la)*(-cs) +
		       mu*(-cs)+(2*mu+la)*(-cs) - cs/k2;
	       }
               else if( nr == 5 )
	       {
                  double sc = sin(x[0])*cos(x[1]);
  	          double ss = sin(x[0])*sin(x[1]);
                  double cs = cos(x[0])*sin(x[1]);
		  double cc = cos(x[0])*cos(x[1]);
		  double fi, fix, fiy, fixx, fixy, fiyy;
                  fobj.fivals( x, fi, fix, fiy, fixx, fixy, fiyy );
                  ex[dex](vofs[vi],0) = (fixx*sc+2*fix*cc-fi*sc)*(2*mu+la) +
		       mu*(fiyy*sc-2*fiy*ss-fi*sc)+(la+mu)*(fixy*cs+fix*cc-fiy*ss-fi*sc);
                  ex[dex](vofs[vi],1) = (mu+la)*(fixy*sc-fix*ss+fiy*cc-fi*cs) +
		       mu*(fixx*cs-2*fix*ss-fi*cs)+(2*mu+la)*(fiyy*cs+2*fiy*cc-fi*cs);
	       }
               else if( nr == 6 )
	       {
		  Real r = x[0];
                  double sc = sin(x[0])*cos(x[1]);
  	          double ss = sin(x[0])*sin(x[1]);
                  double cs = cos(x[0])*sin(x[1]);
		  double cc = cos(x[0])*cos(x[1]);
                  ex[dex](vofs[vi],0) = sc*(-4*mu- 2*la-(2*mu+la)/(r*r)) +
		        (2*mu+la)/r*cc - sc/k2;
                  ex[dex](vofs[vi],1) = (-4*mu-2*la)*cs - (2*mu+la)/r*ss - cs/k2;
	       }
               else if( nr == 7 )
	       {
		  double cx = cos(x[0])*cos(x[0]);
		  double sx = sin(x[0])*sin(x[0]);
		  double cy = cos(x[1])*cos(x[1]);
		  double sy = sin(x[1])*sin(x[1]);
                  ex[dex](vofs[vi],0) =  2*(mu+la)*( (cx-sx)*cy+cx*(cy-sy))-
		    2*mu*( (cx-sx)*sy+sx*(cy-sy));
	       }
               else if( nr == 8 )
	       {
                  double sc = sin(x[0])*cos(x[1]);
  	          double ss = sin(x[0])*sin(x[1]);
                  double cs = cos(x[0])*sin(x[1]);
		  double cc = cos(x[0])*cos(x[1]);
		  Real r = x[0];
                  ex[dex](vofs[vi],0) = -( mu*(3*r+1)+la*(r+1) )*sc +
		              3*(2*mu+la)*cc - r*sc/k2;
                  ex[dex](vofs[vi],1) = -(mu*(r+3)+la*(r+1))*cs -2*(mu+la)*ss-cs/k2;
                  if( r > 0 )
		    ex[dex](vofs[vi],1) -= mu*ss/r;
		  else
		    ex[dex](vofs[vi],1) -= mu*sin(x[1]);
	       }
               else if( nr == 9 )
	       {
		  if( fabs(x[0])>1e-10 )
                     ex[dex](vofs[vi],0) = mu*( (1/x[0]-2*x[0])*sin(x[0])*cos(x[1])
                               +3*cos(x[0])*cos(x[1]) ) -k2*x[0]*sin(x[0])*cos(x[1]);
                  else
                     ex[dex](vofs[vi],0) = mu*( cos(x[1])
                               +3*cos(x[0])*cos(x[1]) ) -k2*x[0]*sin(x[0])*cos(x[1]);
	       }
               else if( nr == 10 )
	       {
		  if( fabs(x[0])<1e-10 )
		     x[0] = 1e-10;
		  ex[dex](vofs[vi],0) = testrzwk( x[0], x[1], mu, la );
	       }
               if( ebis[dex].isIrregular(b()) && divmul == -1 )
	       {
                 for( int c=0 ; c < nc ; c++ )
   		    ex[dex](vofs[vi],c) /= ebis[dex].volFrac(vofs[vi]);
	       }
               else if( ebis[dex].isIrregular(b()) && divmul == 1 )
	       {
                 for( int c=0 ; c < nc ; c++ )
		    ex[dex](vofs[vi],c) *= ebis[dex].volFrac(vofs[vi]);
	       }
	    }
	 }
   }
}


//-----------------------------------------------------------------------
void divide_volume( LevelData<EBCellFAB>& u, const EBISLayout& ebis )
{
   int nc = u.nComp();
   for( DataIterator li=ebis.getDisjointLayout().dataIterator() ; li.ok() ; ++li )
   {
      DataIndex dex(li());
      for( BoxIterator b(u[dex].getRegion()); b.ok(); ++b )
	 if( !ebis[dex].isCovered(b()) )
	 {
	    Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
            for( int vi=0 ; vi < vofs.size() ; vi++ )
	    {
               if( vofs.size() > 1 )
		  cout << "voff voff" << endl;
               for( int c = 0 ; c < nc; c++ )
                  u[dex](vofs[vi],c) /= ebis[dex].volFrac(vofs[vi]);
	    }
	 }
   }
}


//-----------------------------------------------------------------------
void multiply_volume( LevelData<EBCellFAB>& u, const EBISLayout& ebis )
{
   int nc = u.nComp();
   for( DataIterator li=ebis.getDisjointLayout().dataIterator() ; li.ok() ; ++li )
   {
      DataIndex dex(li());
      for( BoxIterator b(u[dex].getRegion()); b.ok(); ++b )
	 if( !ebis[dex].isCovered(b()) )
	 {
	    Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
            for( int vi=0 ; vi < vofs.size() ; vi++ )
	    {
               if( vofs.size() > 1 )
		  cout << "voff voff" << endl;
               for( int c = 0 ; c < nc; c++ )
                  u[dex](vofs[vi],c) *= ebis[dex].volFrac(vofs[vi]);
	    }
	 }
   }
}


//-----------------------------------------------------------------------
void output_array( LevelData<EBCellFAB>& u, const EBISLayout& ebis, char* file )
{
   double *ar;
   int nc = u.nComp();
   int nboxes = ebis.getDisjointLayout().size();
   int fd=open(file, O_CREAT | O_TRUNC | O_WRONLY, 0660 );   
   write( fd, &nboxes, sizeof(int) );

   for( DataIterator li=ebis.getDisjointLayout().dataIterator() ; li.ok() ; ++li )
   {
      DataIndex dex(li());
      IntVect dims = u[dex].getRegion().size();
      IntVect low  = u[dex].getRegion().smallEnd();

      int ni = dims[0];
      int nj = dims[1];
      ar = new double[ni*nj*nc];
      for( BoxIterator b(u[dex].getRegion()); b.ok(); ++b )
      {
        int i=b()[0]-low[0], j=b()[1]-low[1];
	if( !ebis[dex].isCovered(b()) )
	{
	   Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
           for( int vi=0 ; vi < vofs.size() ; vi++ )
	   {
	      for( int c= 0 ; c<nc ; c++ )
	         ar[c+i*nc+nc*ni*j] = u[dex](vofs[vi],c);
	   }
        }
	else
	   for( int c= 0 ; c<nc ; c++ )
	     ar[c+i*nc+nc*ni*j] = 0;
      }
      write(fd,&nc,sizeof(int));
      write(fd,&ni,sizeof(int));
      write(fd,&nj,sizeof(int));
      write(fd,ar,sizeof(double)*ni*nj*nc);
      delete[] ar;
   }
   close(fd);
}


//-----------------------------------------------------------------------
void output_level_array( const Vector<LevelData<EBCellFAB>*>& u,
 		         const Vector<EBISLayout>& ebis,
		         const Vector<RealVect> dx,
		         const RealVect x0, char* file )

{
   int fd=open(file, O_CREAT | O_TRUNC | O_WRONLY, 0660 );   
   int nlevels = u.size();
   write(fd,&nlevels,sizeof(int));
   write(fd,&x0[0],sizeof(double));
   write(fd,&x0[1],sizeof(double));
   for( int l=0 ; l < nlevels ; l++ )
   {
      write(fd,&dx[l][0],sizeof(double));
      write(fd,&dx[l][1],sizeof(double));
   }
   //   cout << "output_level_array" << endl;

   for( int l=0 ; l < nlevels ; l++ )
   {
     //      cout << "level " << l << endl;
      int nc = u[l]->nComp();
      for( DataIterator li=ebis[l].getDisjointLayout().dataIterator() ; li.ok() ; ++li )
      {
         DataIndex dex(li());
         IntVect dims = (*u[l])[dex].getRegion().size();
         IntVect low  = (*u[l])[dex].getRegion().smallEnd();
         int ni = dims[0];
         int nj = dims[1];
         int *flag    = new int[ni*nj];
         double *ar   = new double[ni*nj*nc];
	 //         cout << "past malloc " << endl;
         for( BoxIterator b((*u[l])[dex].getRegion()); b.ok(); ++b )
         {
            int i=b()[0]-low[0], j=b()[1]-low[1];
	    Vector<VolIndex> vofs = ebis[l][dex].getVoFs(b());
            for( int vi=0 ; vi < vofs.size() ; vi++ )
	    {
	       for( int c= 0 ; c<nc ; c++ )
	          ar[c+i*nc+nc*ni*j] = (*u[l])[dex](vofs[vi],c);
		  //		  cout << file << " " << i << "," << j << " " << c << (*u[l])[dex](vofs[vi],c)<<endl;}
	       //               cout << "volFrac " << b() << " " << ebis[l][dex].volFrac(vofs[vi]) << endl;
	    }
   	    if( ebis[l][dex].isCovered(b()) )
	       flag[i+ni*j] = 0;
	    else if( ebis[l][dex].isRegular(b()) )
	       flag[i+ni*j] = 1;
	    else
	       flag[i+ni*j] = 2;
	 }
	 //	 cout << "writing..." << endl;
         write(fd,&l, sizeof(int));
         write(fd,&nc,sizeof(int));
         write(fd,&ni,sizeof(int));
         write(fd,&nj,sizeof(int));
         write(fd,&low[0],sizeof(int));
         write(fd,&low[1],sizeof(int));
         write(fd,ar,sizeof(double)*ni*nj*nc);
         write(fd,flag,sizeof(int)*ni*nj);
         delete[] ar;
         delete[] flag;
      }
   }
   close(fd);
}
//-----------------------------------------------------------------------
 void out_shape( RealVect normal,
		 RealVect bndrycentroid,
		 IntVect  ind,
		 RealVect centroid,
		 RealVect dx,
                 ofstream& filut )
{
  RealVect tan;
  tan[0] = -normal[1];
  tan[1] =  normal[0];

  RealVect x, pup, pdn, bcc, ccc;
  x[0] = (ind[0]+0.5)*dx[0];
  x[1] = (ind[1]+0.5)*dx[1];
  pup[0] = (ind[0]+1)*dx[0];
  pup[1] = (ind[1]+1)*dx[1];
  pdn[0] = (ind[0])*dx[0];
  pdn[1] = (ind[1])*dx[1];
  bcc[0] = x[0] + bndrycentroid[0]*dx[0];
  bcc[1] = x[1] + bndrycentroid[1]*dx[1];
  ccc[0] = x[0] + centroid[0]*dx[0];
  ccc[1] = x[1] + centroid[1]*dx[1];
  Vector<Real> slims(4);
  if( tan[0] != 0 )
  {
     slims[0] = (pup[0]-bcc[0])/tan[0];
     slims[1] = (pdn[0]-bcc[0])/tan[0];
  }
  else
  {
    slims[0] = 1e38;
    slims[1] = 1e38;
  }
  if( tan[1] != 0 )
  {
     slims[2] = (pup[1]-bcc[1])/tan[1];
     slims[3] = (pdn[1]-bcc[1])/tan[1];
  }
  else
  {
    slims[2] = 1e38;
    slims[3] = 1e38;
  }
  // find the two with smallest pos and smallest neg

  int i1 = 0, i2=0;

  for( int i=0 ; i<4 ;i++ )
  {
    if( slims[i] >= 0 )
      i1 = i;
    else
      i2 = i;
  }
  for( int i=0 ; i<4 ;i++ )
  {
    if( fabs(slims[i])<fabs(slims[i1]) && slims[i]>=0 )
      i1 = i;
    if( fabs(slims[i])<fabs(slims[i2]) && slims[i]< 0 )
      i2 = i;
  }

  //  if( ind[0] == 1 && ind[1] == 5 )
  //  {
  //    cout << slims[0] << " " << slims[1] << " " << slims[2] << " " << slims[3] << endl;
  //    cout << i1 << " " << i2 << endl;
  //    exit(2);
  //  }

  RealVect p1,p2;
  p1[0] = bcc[0]+tan[0]*slims[i1];
  p1[1] = bcc[1]+tan[1]*slims[i1];
  p2[0] = bcc[0]+tan[0]*slims[i2];
  p2[1] = bcc[1]+tan[1]*slims[i2];
  
  filut << ind[0] << " " << ind[1] << " " << ccc[0] << " " << ccc[1] << " " <<
    bcc[0] << " " << bcc[1] << " " << p1[0] << " " << p1[1] << " " 
      << p2[0] << " " << p2[1] << " " << normal[0] << " " << normal[1] << endl;

}


//-----------------------------------------------------------------------
UnionIF* makeCrossSection(const Vector<Vector<RealVect> >& a_polygons)
{
  // The final result
  UnionIF* retval;

  // Get the number of polygons and make this inside of the domain
  // the inside of the polygons
  int numPolys = a_polygons.size();
  bool inside = true;

  // A list of all the polygons as implicit functions
  Vector<BaseImplicitFunction*> polytopes;
  polytopes.resize(0);

  // Process each polygon
  for (int p = 0; p < numPolys; p++)
  {
    // All the half planes/spaces used to make a polygon
    Vector<BaseImplicitFunction*> planes;
    planes.resize(0);

    // Get the current polygon (as a vector of points)
    const Vector<RealVect>& polygon = a_polygons[p];

    // Get the number of points in the polygon
    int numPts = polygon.size();

    // Process each pair of points
    for (int n = 0; n < numPts; n++)
    {
      // The normal and point is space used to specify each half plane/space
      RealVect normal(RealVect::Zero);
      RealVect point;

      // Set the normal remembering that the last point connects to the first
      // point.
      normal[0] = -(polygon[(n+1) % numPts][1] - polygon[n][1]);
      normal[1] =  (polygon[(n+1) % numPts][0] - polygon[n][0]);

      point = polygon[n];

      // Generate the appropriate half plane/space (as an implicit function)
      PlaneIF* plane;
      plane = new PlaneIF(normal,point,inside);

      // Save the result
      planes.push_back(plane);
    }

    // Intersect all the half planes/spaces to create an implicit function
    // that represents the polygon
    IntersectionIF* polygonIF = new IntersectionIF(planes);

    polytopes.push_back(polygonIF);
  }

  // Union all the polygon implicit functions to get the implicit function
  // returned
  retval = new UnionIF(polytopes);

  return retval;
}




//-----------------------------------------------------------------------
GeometryShop* make_nozzle()
{
  bool     insideRegular;

  Vector<Vector<RealVect> > polygonsBody;
  get_polygons( polygonsBody, "body");

  Vector<Vector<RealVect> > polygonsPoppet;
  get_polygons( polygonsPoppet, "poppet");

  // Make the vector of (convex) polygons (vectors of points) into a union
  // of convex polygons, each made from the intersection of a set of half
  // planes/spaces - all represented by implicit functions.
  Vector<Vector<RealVect> > polygons = polygonsBody;
  polygons.append(polygonsPoppet);

  UnionIF* crossSection = makeCrossSection(polygons);

  ImplicitFunctionBL *myGeom;

  if (SpaceDim == 2)
  {
    // In 2D use "as is"
    ComplementIF insideOut(*crossSection);
    //    retval = insideOut.newImplicitFunction();
    myGeom = new ImplicitFunctionBL(insideOut);
  }
  else
  {
    // In 3D rotate about the z-axis
    LatheIF implicit(*crossSection,insideRegular);
    //    retval = implicit.newImplicitFunction();
    myGeom = new ImplicitFunctionBL(implicit);
  }
  GeometryShop* workshop = new GeometryShop(*myGeom);
  delete myGeom;
  delete crossSection;
  return workshop;
}


//-----------------------------------------------------------------------
void get_polygons( Vector<Vector<RealVect> >& polygons, const char* name )
{
// Get number of polygons
  if( strcmp(name,"poppet") == 0 )
  {
     polygons.resize(1);
     polygons[0].resize(4);
     polygons[0][0] = RealVect(0.8528400,-10.00000000);
     polygons[0][1] = RealVect(0.8521970, -0.32181000);
     polygons[0][2] = RealVect(0.0000000,  0.69393192);
     polygons[0][3] = RealVect(0.0000000, -10.00000000);
  }
  else if( strcmp(name,"body") == 0 )
  {
     polygons.resize(2);

     polygons[0].resize(4);
     polygons[0][0] = RealVect( 20.0000000, 5.64012000);
     polygons[0][1] = RealVect( 1.2499200,  5.64012000);
     polygons[0][2] = RealVect( 1.2499200,  -1.00000000);
     polygons[0][3] = RealVect( 20.0000000,  -1.00000000);

     polygons[1].resize(6);
     polygons[1][0] = RealVect( 1.2499200,   5.64012000 );
     polygons[1][1] = RealVect( 1.1058800,   5.64012000 );
     polygons[1][2] = RealVect( 0.3966470,   1.62330000 );
     polygons[1][3] = RealVect( 0.3964730,   0.42332500 );
     polygons[1][4] = RealVect( 0.7493570,   0.00176610 );
     polygons[1][5] = RealVect( 1.2499200,  -0.00777410 );
  }
}


//-----------------------------------------------------------------------
int main( int argc, char** argv )
{
  int testcase = 4;
  int n_ebis_ghost = 2;
  int n_level_ghost = 1;
  int domainnr = 0;
  int ebgeom   = 0;
  int sol = 1;
  int dirichlet = 1;
  int nit = 1000;

  bool rz = false;
  bool resmulvol = false;
  int ni, nj, nimax, njmax;
  int relax_type = 0, order_eb = 2;   
  int nlevels = 1;
  Real k2, my, la, cof;

  if( argc > 1 )
     read_pars( argv[1], testcase, relax_type, order_eb, ebgeom,
	        domainnr, k2, my, la, rz, ni, nj, nimax, njmax, cof,
                sol, nit, nlevels, n_level_ghost, n_ebis_ghost, dirichlet, resmulvol );
  else
  {
     cout << "Error: No input file specified " << endl;	 
     exit(2);
  }

  EBAMRPoissonOp* amrop;
  EBAMRPoissonOpFactory* amrop_factory;

  EBAMRHelmholtzOp* hamrop;
  EBAMRHelmholtzOpFactory* hamrop_factory;

  EBAMRNSviscOp* visc_amrop;
  EBAMRNSviscOpFactory* visc_amrop_factory;

// Domain [-2,2]x[-2,2] at finest resolution, nimax x njmax pts.
//    ==> Index space: 0 - nimax-1, x(i) = -2 + (i+1/2)*h, with h=4/nimax. 
  Box dom( IntVect(0,0), IntVect(nimax-1,njmax-1) );

  RealVect x0; //origin
  Real dxf;  // h_finest 

  if( domainnr == 1 )
  {
//  [-2,2]x[-2,2]
     x0[0] = -2.0;
     x0[1] = -2.0; 
     dxf   = 4.0/nimax;
  }
  else if( domainnr == 2 )
  {
//  [0.5,4.5]x[0.5,4.5]
     x0[0] = 0.5;
     x0[1] = 0.5;
     dxf   = 4.0/nimax;
  }
  else if( domainnr == 3 )
  {
//  [0,0]x[4,4]
     x0[0] = 0.0;
     x0[1] = 0.0;
     dxf   = 4.0/nimax;
  }
  else if( domainnr == 4 )
  {    
//  [0,-1.5]x[10.0,18.5]
     x0[0] = 0;
     x0[1] = -1.5;
     dxf = 10.0/nimax;
  }
  else if( domainnr == 5 )
  {    
//  [0,-1.5]x[5.0,38.5]
     x0[0] = 0;
     x0[1] = -1.5;
     dxf = 5.0/nimax;
  }

  GeometryShop* workshop;
  if( ebgeom >= 0 && ebgeom <= 4 )
  {
// Add cylinder to domain
     RealVect c0;
     if( ebgeom == 0 )
     {
        c0[0] = 0.0;
        c0[1] = 0.0;
     }
     else if( ebgeom == 1 )
     {
        c0[0] = 2.5;
        c0[1] = 2.5;
     }
     else if( ebgeom == 2 )
     {
        c0[0] = 100;
        c0[1] = 100;
     }
     else if( ebgeom == 3 )
     {
        c0[0] = 1.0;
        c0[1] = x0[1]+4-0.2;
     }
     else if( ebgeom == 4 )
     {
        c0[0] = 1.0;
        c0[1] = x0[1]+4+0.2;
     }
     CylinderBL cyl( 0.5, 0.0, c0, 0, false, dxf ); 
     workshop = new GeometryShop(cyl);
  }
  else if( ebgeom == 5 )
  {
      workshop = make_nozzle();
  }

// define EBIS
   EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
   ebisPtr->define( dom, x0, dxf, *workshop, 64 ); 

  //  int maxcoarse = -1; // 7 coarsenings, means 7 pts on coarsest.
                      //-1 means coarse all the way to one grid point.
  //  int maxcells  = 50; // Max no of cells in a grid (??), not sure what value is reasonable.

   //   ebisPtr->define( dom, x0, dxf, workshop, maxcells, maxcoarse );

   Vector<int> ngrid(nlevels);
   ngrid[0] = 1;
   if( nlevels == 2 )
   {
      ngrid[0] = 1;
      ngrid[1] = 3;
      CH_assert( ni>20 && nj>20 );
   }
   else if( nlevels == 3 )
   {
      ngrid[0] = 1;
      ngrid[1] = 3;
      ngrid[2] = 2;
      CH_assert( ni>20 && nj>20 );
   }

   Vector<int> refrats(nlevels,2);
   
   Vector<DisjointBoxLayout> dbls(nlevels);
   Vector<EBISLayout> ebises(nlevels);
   Vector<RealVect> dx(nlevels);
   Vector<ProblemDomain> doms(nlevels);

// Define grids on all levels
   int p2 = 1;
   cout << ni << " " << nj << endl;
   for( int l=0 ; l < nlevels ; l++ )
   {
      Box domain(IntVect(0,0),IntVect(p2*ni-1,p2*nj-1));

      cout << "level " << l << " domain = " << domain << endl;
      doms[l]    = ProblemDomain( domain );
      if( domainnr == 4 )
         dx[l]      = RealVect(10.0/(p2*ni),20.0/(p2*nj));
      else if( domainnr == 5 )
         dx[l]      = RealVect(5.0/(p2*ni),40.0/(p2*nj));
      else
         dx[l]      = RealVect(4.0/(p2*ni),4.0/(p2*nj));

      p2 = 2*p2;

      Vector<int> prnrs(ngrid[l],0);
      Vector<Box> boxes(ngrid[l]);
      if( l == 0 )
      {
	 boxes[0] = domain;
      }
      else if( l == 1 )
      {
// Note, to have properly nested with ref=2, begin in even end in odd
 	 boxes[0] = Box( IntVect(4,4),IntVect(11,21) );
	 boxes[1] = Box( IntVect(14,4), IntVect(21,15) );
	 boxes[2] = Box( IntVect(26,2), IntVect(33,33) );
      }
      else if( l == 2 )
      {
// Note, to have properly nested with ref=2, begin in 4*s end in 4s-1
 	 boxes[0] = Box( IntVect(8,8),IntVect(43,27) );
	 boxes[1] = Box( IntVect(64,16), IntVect(103,31) );
      }
      //      dbls[l] = DisjointBoxLayout( boxes, prnrs );
      dbls[l].define( boxes, prnrs );
      ebisPtr->fillEBISLayout( ebises[l], dbls[l], domain, n_ebis_ghost );
      dbls[l].print();
      dbls[l].close();
   }


// Set boundary condition on domain boundary.
   RefCountedPtr<BaseBCVectValue> bc_function;
   if( rz && dirichlet == 1 )
      bc_function = new DirichletVectFunctionHHRZ();
   else if( dirichlet == 1 )
      bc_function = new DirichletVectFunctionHH();
   else if( rz )
      bc_function = new NeumannVectFunctionHHRZ();
   else
      bc_function = new NeumannVectFunctionHH();

   HelmholtzDomainBCFactory* bc_hh;
   if( dirichlet == 1 )
      bc_hh = new DirichletHelmholtzDomainBCFactory();
   else
      bc_hh = new NeumannHelmholtzDomainBCFactory();
   bc_hh->set_function( bc_function );
   bc_hh->set_rz( rz );
   bc_hh->set_order( 2 );
   RefCountedPtr<HelmholtzDomainBCFactory> bc_factory = bc_hh;


// Set boundary condition on embedded boundary.
   HelmholtzEBBCFactory* ebbc_hh;
   if( dirichlet == 1 )
      ebbc_hh = new DirichletHelmholtzEBBCFactory();
   else
      ebbc_hh = new NeumannHelmholtzEBBCFactory();
   ebbc_hh->set_function( bc_function );
   ebbc_hh->set_rz( rz );
   ebbc_hh->set_order( 2 );
   RefCountedPtr<HelmholtzEBBCFactory> ebbc_factory = ebbc_hh;

   RefCountedPtr<BaseDomainBCFactory> bc_factory_poisson;
   RefCountedPtr<BaseEBBCFactory> bceb_factory_poisson;

   if( testcase == 1 )
   {
       BaseBCValue* bc_func_poisson = new DirichletFunction();
       BaseBCFlux* bc_flux_poisson = new NeumannFunction();
      if( dirichlet == 1 )
      {

          DirichletPoissonDomainBCFactory* bc_dir = new DirichletPoissonDomainBCFactory();
          bc_dir->setFunction( bc_func_poisson );
          bc_factory_poisson = bc_dir;
      }
      else
      {
         NeumannPoissonDomainBCFactory* bc_dir = new NeumannPoissonDomainBCFactory();
         bc_dir->setFunction( bc_flux_poisson );
         bc_factory_poisson = bc_dir;
      }
     if( dirichlet == 1 )
     {
        DirichletPoissonEBBCFactory *ebbc_dir = new DirichletPoissonEBBCFactory();
        ebbc_dir->setOrder( 2 );
        ebbc_dir->setFunction( bc_func_poisson );
        bceb_factory_poisson = ebbc_dir;
     }
     else
     {
        NeumannPoissonEBBCFactory *ebbc_dir = new NeumannPoissonEBBCFactory();
        ebbc_dir->setFunction( bc_flux_poisson );
        bceb_factory_poisson = ebbc_dir;
     }
   }

   Vector<LevelData<EBCellFAB>*> rhos(nlevels);
   int rhodata = 1;
   for( int l = 0 ; l < nlevels ; l++ )
   {
      EBCellFactory cfactory( ebises[l] );
      rhos[l] = new LevelData<EBCellFAB>( dbls[l], 1, IntVect(1,1), cfactory );
      get_rho( *rhos[l], ebises[l], dx[l], x0, rhodata );
      //      rhos[l] = new LevelData<EBCellFAB>(dbls[l], 1, IntVect::Zero, cfactory );
      //      rhos[l]->define( dbls[l], 1, IntVect::Zero, cfactory );
      //      for( DataIterator li=dbls[l].dataIterator() ; li.ok() ; ++li )
      //          (*rhos[l])[li()].setVal(1.0);
   }

   IntVect dum1=IntVect::Unit, dum2=IntVect::Zero;
   amrop_factory  = new EBAMRPoissonOpFactory( dbls, ebises, refrats, doms[0], dx[0],
					      x0, order_eb, relax_type, bc_factory_poisson,
					      bceb_factory_poisson,k2,my,dum1,dum2);

   hamrop_factory = new EBAMRHelmholtzOpFactory( dbls, ebises, refrats, doms[0], dx[0],
					      x0, k2, order_eb, relax_type, bc_factory,
					      ebbc_factory, my, rhos, rz );

   RefCountedPtr<BaseBCVectValue> bc_function2;
   if( testcase == 8 || testcase == 9 || testcase == 13 )
     bc_function2 = new DirichletRZVectFunction();
   else
     bc_function2 = new DirichletVectFunction();

   DirichletNSviscDomainBCFactory* bc_dir2 = new DirichletNSviscDomainBCFactory();

   bc_dir2->setFunction( bc_function2 );
   //   bc_dir2->setValue(0.0);
   RefCountedPtr<DirichletNSviscDomainBCFactory> bc_factory2 = bc_dir2;

   DirichletNSviscEBBCFactory *ebbc_dir2 = new DirichletNSviscEBBCFactory();
   ebbc_dir2->setOrder( 1 );
   ebbc_dir2->setFunction( bc_function2 );
   //   ebbc_dir2->setValue(0.0);
   RefCountedPtr<DirichletNSviscEBBCFactory> ebbc_factory2=ebbc_dir2;

   Vector<Real> cofv(nlevels,cof);
   visc_amrop_factory = new EBAMRNSviscOpFactory( dbls, ebises, refrats, doms[0], dx[0],
					          x0, order_eb, relax_type, bc_factory2,
				          ebbc_factory2, cofv, my, la, rhos, rz );

   Vector<LevelData<EBCellFAB>*> u(nlevels), rhs(nlevels), res(nlevels), err(nlevels);  

   int ncomp = 1;
   int rescomp = 1;
   if( testcase == 6 || testcase == 7 || testcase == 8 || testcase == 9 ||
       testcase == -1 || testcase == -2 )
   {
      ncomp = 2;
      rescomp = 2;
   }
   if( testcase == 12 || testcase == 13 )
   {
      ncomp = 2;
      rescomp = 1;
   }

   for( int l = 0; l < nlevels ; l++ )
   {
      EBCellFactory cfactory( ebises[l] );
      IntVect ghosts(n_level_ghost,n_level_ghost);
      u[l]   = new LevelData<EBCellFAB>( dbls[l], ncomp, ghosts, cfactory ); 
      rhs[l] = new LevelData<EBCellFAB>( dbls[l], rescomp, IntVect::Zero, cfactory );
      if( testcase == 1 || testcase == 4 || testcase == 10
	  || testcase == 6 || testcase == 8 || testcase == 12 || testcase == 13 )
      {
         res[l] = new LevelData<EBCellFAB>( dbls[l], rescomp, IntVect::Zero, cfactory );
         err[l] = new LevelData<EBCellFAB>( dbls[l], rescomp, IntVect::Zero, cfactory );
      }
      else
      {
         res[l] = new LevelData<EBCellFAB>( dbls[l], ncomp, ghosts, cfactory );
         err[l] = new LevelData<EBCellFAB>( dbls[l], ncomp, ghosts, cfactory );
      }
   }

   if( testcase == -3 )
   {
       cout << "EBIS levels: " << ebisPtr->numLevels() << endl;
       int npts=0;
       cout << "Level test: Give npts " << endl;
       cin >> npts;
       while( npts != -1 )
       {
          Box tbox( IntVect(0,0), IntVect(npts-1,npts-1) );
          cout << "Level to this index: " << ebisPtr->getLevel( tbox ) << endl;
          cout << "Level test: Give npts " << endl;
          cin >> npts;
       }
   }
   else if( testcase == 1 )
   {
// Solve Helmholtz problem for a known forcing
      cout << "------------------------------------------------------------\n";
      cout << "Test 1, Poisson equation, with solution sin(x)*cos(y).\n" <<
	      "Solving the equation." << endl;
      cout << "------------------------------------------------------------\n";

      for( int l = 0 ; l < nlevels ; l++ )
      {
// Give forcing:
         EBLevelDataOps::setToZero(*rhs[l]);
         get_exactres( *rhs[l], ebises[l], dx[l], x0, 2, 1, k2, my, la );    
// Initial guess:
         EBLevelDataOps::setToZero(*u[l]);
      }

      if( sol == 1 )
      {
// Set up MG solver
         AMRMultiGrid<LevelData<EBCellFAB> > mgsolver;
         BiCGStabSolver<LevelData<EBCellFAB> > bottomSolver;      
         int maxlevels = nlevels;
         mgsolver.define( doms[0], *amrop_factory, &bottomSolver, maxlevels );
         int l_max = nlevels-1, l_base = 0;
	 //         mgsolver.solve(  uv2, rhsv2, l_max, l_base );
         mgsolver.solve(  u, rhs, l_max, l_base );
      }
      //      else if( sol == 2 )
      //      {
      //	 for( int l = 0 ; l < nlevels ; l++ )
      //	 {
      //            amrop = amrop_factory->createOperator( dbls[l], ebises[l], doms[l],
      //						     dx[l], l );
      //	    amrop->set_inhomogeneous_relax();
      //	    amrop->relax( *u[l], *rhs[l], nit );
      //	 }
      //      }
      output_level_array( u, ebises, dx, x0, "ul.bin" );

// Get exact solution and compare
      for( int l = 0 ; l < nlevels ; l++ )
         get_testdata( *res[l], ebises[l], dx[l], x0, 1 );

      output_level_array( res, ebises, dx, x0, "exl.bin");

      for( int l=0 ; l< nlevels ; l++ )
      {
         EBLevelDataOps::scale(*res[l],-1.0);
         EBLevelDataOps::sum( *err[l], *u[l], *res[l] );
         amrop = amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Level " << l << " Error norm L2: " << amrop->norm(*err[l],2);
         cout << "Level " << l << "Error norm Linf: " << amrop->norm(*err[l],0) << endl;
      }
   }
   else if( testcase == 3 )
   {
// Solve Helmholtz problem for a known forcing
      cout << "------------------------------------------------------------\n";
      cout << "Test 3, Helmholtz' equation, with solution sin(x)*cos(y).\n" <<
	      "Solving the equation." << endl;
      cout << "------------------------------------------------------------\n";

      for( int l = 0 ; l < nlevels ; l++ )
      {
// Give forcing:
         EBLevelDataOps::setToZero(*rhs[l]);
         get_exactres( *rhs[l], ebises[l], dx[l], x0, 2, 1, k2, my, la );    
// Initial guess:
         EBLevelDataOps::setToZero(*u[l]);
      }

      if( sol == 1 )
      {
// Set up MG solver
         AMRMultiGrid<LevelData<EBCellFAB> > mgsolver;
         BiCGStabSolver<LevelData<EBCellFAB> > bottomSolver;      
         int maxlevels = nlevels;
         mgsolver.define( doms[0], *hamrop_factory, &bottomSolver, maxlevels );
         int l_max = nlevels-1, l_base = 0;
	 //         mgsolver.solve(  uv2, rhsv2, l_max, l_base );
         mgsolver.solve(  u, rhs, l_max, l_base );
      }
      else if( sol == 2 )
      {
	 for( int l = 0 ; l < nlevels ; l++ )
	 {
            hamrop = hamrop_factory->createOperator( dbls[l], ebises[l], doms[l],
						     dx[l], l );
	    hamrop->set_inhomogeneous_relax();
	    hamrop->relax( *u[l], *rhs[l], nit );
	 }
      }
      output_level_array( u, ebises, dx, x0, "ul.bin" );

// Get exact solution and compare
      for( int l = 0 ; l < nlevels ; l++ )
         get_testdata( *res[l], ebises[l], dx[l], x0, 1 );

      output_level_array( res, ebises, dx, x0, "exl.bin");

      for( int l=0 ; l< nlevels ; l++ )
      {
         EBLevelDataOps::scale(*res[l],-1.0);
         EBLevelDataOps::sum( *err[l], *u[l], *res[l] );
         hamrop = hamrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Level " << l << " Error norm L2: " << hamrop->norm(*err[l],2);
         cout << "Level " << l << "Error norm Linf: " << hamrop->norm(*err[l],0) << endl;
      }
   }
   else if( testcase == 4 )
   {
// Evaluate Helmholtz operator on known function, print out trunc errors.
      cout << "------------------------------------------------------------\n";
      cout << "Test 4, Helmholtz equation, check truncation error.\n" <<
	      "Evaluate operator and compare trunc error with its known value." << endl;
      cout << "------------------------------------------------------------\n";

// Get test data
      for( int l = 0; l < nlevels ; l++ )
         get_testdata( *u[l], ebises[l], dx[l], x0, 1 );
      //         get_testdata( *u[l], ebises[l], dx[l], x0, 7 );

// Apply operator
      for( int l = 0 ; l < nlevels ; l++ )
      {
         hamrop = hamrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         if( l == 0 )
            hamrop->applyOp( *res[l], *u[l], NULL, false, false, NULL );
	 else
            hamrop->applyOp( *res[l], *u[l], u[l-1], false, false, NULL );
         if( l < nlevels-1 )
	    hamrop->reflux( *res[l], *u[l+1], *u[l] ); 

         get_exactres( *rhs[l], ebises[l], dx[l], x0, 2, 0, k2, my, la );

	 if( resmulvol )
	    multiply_volume( *rhs[l], ebises[l] );
	 else
            divide_volume( *res[l], ebises[l] );
      }
// Get exact residual and compare
      output_level_array( res, ebises, dx, x0, "resl.bin");
      output_level_array( rhs, ebises, dx, x0, "exl.bin" );

      for( int l = 0 ; l < nlevels ; l++ )
      {
         EBLevelDataOps::scale(*rhs[l],-1.0);
         EBLevelDataOps::sum( *err[l], *rhs[l], *res[l] );
         hamrop = hamrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Error norm L2: " << hamrop->norm(*err[l],2);
         cout << "Error norm Linf: " << hamrop->norm(*err[l],0) << endl;
      }
   }
   else if( testcase == 5 )
   {
      cout << "------------------------------------------------------------\n";
      cout << "Test 5, Misc box/ebisbox tests.\n" <<
      cout << "------------------------------------------------------------\n";

      Vector<int> prs(3,0);
      Vector<Box> testdoms(3);
      testdoms[0] = Box( IntVect(0,0), IntVect(10,12) );
      testdoms[1] = Box( IntVect(15,13),IntVect(20,20) );
      testdoms[2] = Box( IntVect(23,20), IntVect(40,32) );

      DisjointBoxLayout dl( testdoms, prs );
      EBISLayout el;
      int nghost;
      cout << "Give nghost for ebis layout " << endl;
      cin >> nghost;
      cout << "Give nghost for level data " << endl;
      int nld;
      cin >> nld;
      IntVect gh(nld,nld);

      ebisPtr->fillEBISLayout( el, dl, Box(IntVect(0,0),IntVect(ni-1,nj-1)), nghost );
      EBCellFactory cfactory( el );

      LevelData<EBCellFAB>  w( dl, 1, gh, cfactory ); 

      for( DataIterator li=dl.dataIterator() ; li.ok() ; ++li )
      {

         DataIndex dex(li());
	 cout << "-----------------" << endl;
         cout << "EBisbox region = " << el[dex].getRegion() << endl;
         cout << "EBisbox domain = " << el[dex].getDomain() << endl;
         cout << "u region = "  << w[dex].getRegion() << endl;
         cout << "u box    = "  << w[dex].box() << endl;
         IntVectSet ivs = el[dex].getIrregIVS( el[dex].getRegion() );
         IntVectSet ivs2= el[dex].getIrregIVS( dl.get(li()) );
         cout << "Ivs is: " << endl;
         cout << "Box size is " << w[dex].box().size() << endl;
         IntVectSet ivtest;
	 ivtest |= IntVect(4,4);
	 ivtest |= IntVect(4,5);
	 ivtest |= IntVect(4,6);
         BaseIVFAB<Real> bf( ivtest, el[dex].getEBGraph(),1 );
         int nr=0;
	 for( IVSIterator it(ivtest) ; it.ok() ; ++it)
	 {
	    Vector<VolIndex> vof=el[dex].getVoFs(it());
            if( vof.size()>0)
   	      bf(vof[0],0) = nr++;
	 }
         cout << "My own IVSfab : ";
	 for( IVSIterator it(ivtest) ; it.ok() ; ++it)
	 {
	    Vector<VolIndex> vof=el[dex].getVoFs(it());
            if( vof.size()>0)
	    cout << " " << bf(vof[0],0);
	 }
         cout << endl;

         for( IVSIterator it(ivs) ; it.ok() ; ++it)
         {
	    cout << it() ;
            if( el[dex].isCovered(it()) )
 	       cout << " covered";
	    else if( el[dex].isIrregular(it()) )
 	       cout << " irregular";
	    else
	       cout << " regular ";
	    cout << endl;
         }

         cout << "Ivs2 is: " << endl;
         for( IVSIterator it(ivs2) ; it.ok() ; ++it)
         {
	    cout << it() ;
            if( el[dex].isCovered(it()) )
 	       cout << " covered";
	    else if( el[dex].isIrregular(it()) )
	    {
 	       cout << " irregular";
	       Vector<VolIndex> vi = el[dex].getVoFs( it() );
               if( vi.size() > 0 )
	       {
                  cout << " Bndry normal is " << el[dex].normal(vi[0]) << " ";
		  cout << " Vol frac is " << el[dex].volFrac(vi[0]);
	       }
	       if( vi.size() > 1 )
		 cout << "voff voff";
	    }
	    else
	       cout << " regular ";
	    cout << endl;
	 }
      }
   }
   else if( testcase == 6 )
   {
// Evaluate viscous operator on known function, print out trunc errors.
      cout << "------------------------------------------------------------\n";
      cout << "Test 6, N-S viscosity, check truncation error.\n" <<
	      "Evaluate operator and compare trunc error with its known value." << endl;
      cout << "------------------------------------------------------------\n";

// Get test data
      for( int l = 0; l < nlevels ; l++ )
         get_testdata( *u[l], ebises[l], dx[l], x0, 2 );

// Apply operator
      for( int l = 0 ; l < nlevels ; l++ )
      {
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         if( l == 0 )
            visc_amrop->applyOp( *res[l], *u[l], NULL, false, true, NULL );
	 else
            visc_amrop->applyOp( *res[l], *u[l], u[l-1], false, false, NULL );
         if( l < nlevels-1 )
	    visc_amrop->reflux( *res[l], *u[l+1], *u[l], 0 ); 

         get_exactres( *rhs[l], ebises[l], dx[l], x0, 4, 0, cof, my, la );
	 if( resmulvol )
	    multiply_volume( *rhs[l], ebises[l] );
	 else
            divide_volume( *res[l], ebises[l] );

      }
// Get exact residual and compare
      output_level_array( res, ebises, dx, x0, "resl.bin");
      output_level_array( rhs, ebises, dx, x0, "exl.bin" );

      for( int l = 0 ; l < nlevels ; l++ )
      {
         EBLevelDataOps::scale(*rhs[l],-1.0);
         EBLevelDataOps::sum( *err[l], *rhs[l], *res[l] );
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Error norm L2: " << visc_amrop->norm(*err[l],2);
         cout << "Error norm Linf: " << visc_amrop->norm(*err[l],0) << endl;
      }
   }
   else if( testcase == 7 )
   {
// Solve NS problem for a known forcing
      cout << "------------------------------------------------------------\n";
      cout << "Test 7, N-S viscosity equation, with solution sin(x)*cos(y).\n" <<
	      "Solving the equation." << endl;
      cout << "------------------------------------------------------------\n";

      for( int l = 0 ; l < nlevels ; l++ )
      {
// Give forcing:
         EBLevelDataOps::setToZero(*rhs[l]);
         get_exactres( *rhs[l], ebises[l], dx[l], x0, 4, 1, cof, my, la );    
// Initial guess:
         EBLevelDataOps::setToZero(*u[l]);
      }

      if( sol == 1 )
      {
// Set up MG solver
         AMRMultiGrid<LevelData<EBCellFAB> > mgsolver;
         BiCGStabSolver<LevelData<EBCellFAB> > bottomSolver;      
         int maxlevels = nlevels;
         mgsolver.define( doms[0], *visc_amrop_factory, &bottomSolver, maxlevels );
         int l_max = nlevels-1, l_base = 0;
         mgsolver.solve(  u, rhs, l_max, l_base );
      }
      else if( sol == 2 )
      {
	 for( int l = 0 ; l < nlevels ; l++ )
	 {
            visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l],
						     dx[l], l );
	    visc_amrop->set_inhomogeneous_relax();
	    visc_amrop->relax( *u[l], *rhs[l], nit );
	 }
      }
      output_level_array( u, ebises, dx, x0, "ul.bin" );

// Get exact solution and compare
      for( int l = 0 ; l < nlevels ; l++ )
         get_testdata( *res[l], ebises[l], dx[l], x0, 2 );

      output_level_array( res, ebises, dx, x0, "exl.bin");
      for( int l=0 ; l< nlevels ; l++ )
      {
         EBLevelDataOps::scale(*res[l],-1.0);
         EBLevelDataOps::sum( *err[l], *u[l], *res[l] );
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Level " << l << " Error norm L2:   " << visc_amrop->norm(*err[l],2);
         cout << "Level " << l << " Error norm Linf: " << visc_amrop->norm(*err[l],0) << endl;
      }
   }     
   else if( testcase == 8 )
   {
// Evaluate viscous operator on known function, print out trunc errors, cyl NS.
      cout << "------------------------------------------------------------\n";
      cout << "Test 8, N-S viscosity, check truncation error, cyl coords.\n" <<
	      "Evaluate operator and compare trunc error with its known value." << endl;
      cout << "------------------------------------------------------------\n";

// Get test data
      for( int l = 0; l < nlevels ; l++ )
         get_testdata( *u[l], ebises[l], dx[l], x0, 4 );

// Apply operator
      for( int l = 0 ; l < nlevels ; l++ )
      {
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         if( l == 0 )
            visc_amrop->applyOp( *res[l], *u[l], NULL,   false, true,  NULL );
	 else
            visc_amrop->applyOp( *res[l], *u[l], u[l-1], false, false, NULL );
	 //         if( l < nlevels-1 )
	 //	    visc_amrop->reflux( *res[l], *u[l+1], *u[l], 0 ); 

         get_exactres( *rhs[l], ebises[l], dx[l], x0, 8, 0, cof, my, la );
	 if( resmulvol )
	    multiply_volume( *rhs[l], ebises[l] );
	 else
            divide_volume( *res[l], ebises[l] );
      }
// Get exact residual and compare
      output_level_array( u,   ebises, dx, x0, "ul.bin"  );
      output_level_array( res, ebises, dx, x0, "resl.bin");
      output_level_array( rhs, ebises, dx, x0, "exl.bin" );


      for( int l = 0 ; l < nlevels ; l++ )
      {
         EBLevelDataOps::scale(*rhs[l],-1.0);
         EBLevelDataOps::sum( *err[l], *rhs[l], *res[l] );
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Error norm L2: " << visc_amrop->norm(*err[l],2);
         cout << "Error norm Linf: " << visc_amrop->norm(*err[l],0) << endl;
      }
   }
   else if( testcase == 9 )
   {
// Solve NS problem for a known forcing
      cout << "------------------------------------------------------------\n";
      cout << "Test 9, N-S viscosity equation, with solution sin(x)*cos(y).\n" <<
	      "Solving the equation in cylindrical coordinates." << endl;
      cout << "------------------------------------------------------------\n";

      for( int l = 0 ; l < nlevels ; l++ )
      {
// Give forcing:
         EBLevelDataOps::setToZero(*rhs[l]);
         get_exactres( *rhs[l], ebises[l], dx[l], x0, 8, 1, cof, my, la );
// Initial guess:
         EBLevelDataOps::setToZero(*u[l]);
      }

      if( sol == 1 )
      {
// Set up MG solver
         AMRMultiGrid<LevelData<EBCellFAB> > mgsolver;
         BiCGStabSolver<LevelData<EBCellFAB> > bottomSolver;      
	 //         EBSimpleSolver bottomSolver;      
         int maxlevels = nlevels;
         mgsolver.define( doms[0], *visc_amrop_factory, &bottomSolver, maxlevels );
         int l_max = nlevels-1, l_base = 0;
         mgsolver.solve(  u, rhs, l_max, l_base );
      }
      else if( sol == 2 )
      {
	 for( int l = 0 ; l < nlevels ; l++ )
	 {
            visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l],
						     dx[l], l );
	    visc_amrop->set_inhomogeneous_relax();
	    visc_amrop->relax( *u[l], *rhs[l], nit );
	 }
      }
      output_level_array( u, ebises, dx, x0, "ul.bin" );

// Get exact solution and compare
      for( int l = 0 ; l < nlevels ; l++ )
         get_testdata( *res[l], ebises[l], dx[l], x0, 4 );

      output_level_array( res, ebises, dx, x0, "exl.bin");
      for( int l=0 ; l< nlevels ; l++ )
      {
         EBLevelDataOps::scale(*res[l],-1.0);
         EBLevelDataOps::sum( *err[l], *u[l], *res[l] );
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Level " << l << " Error norm L2:   " << visc_amrop->norm(*err[l],2);
         cout << "Level " << l << " Error norm Linf: " << visc_amrop->norm(*err[l],0) << endl;
      }
   }     
   else if( testcase == -1 || testcase == -2 )
   {
      int l = 0;
      visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
      EBCellFactory cfactory( ebises[l] );

      LevelData<EBCellFAB>   uus( dbls[l], 2, IntVect::Zero, cfactory );
      LevelData<EBCellFAB>   res( dbls[l], 2, IntVect::Zero, cfactory );
      LevelData<EBCellFAB>   wgh( dbls[l], 2, IntVect::Zero, cfactory );
      cout << "Testing diagonal weight, input component 1 to " << SpaceDim << " > ";
      int c;
      cin >> c;

      const EBISLayout& ebis = ebises[l];

      for( DataIterator li=dbls[l].dataIterator() ; li.ok() ; ++li )
      {
         DataIndex dex(li());
         visc_amrop->get_uweight( dex, wgh );
         for( BoxIterator b(ebis[dex].getRegion()); b.ok(); ++b )
         {
            if( !ebis[dex].isCovered(b()) )
	    {
	       Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
               for( int vi=0 ; vi < vofs.size() ; vi++ )
  	       {
                  res[dex].setVal(0.0);
		  uus[dex].setVal(0.0);
		  uus[dex](vofs[vi],c-1) = 1;
		  visc_amrop->applyOp( res, uus, NULL, true, true, NULL );
                  cout << "( " << b()[0] <<" , "<< b()[1] << " ) ";
                  if( ebis[dex].isRegular(b()) )
                    cout << "r" ;
		  else
		    cout << "i" ;
                  Real dif = res[dex](vofs[vi],c-1)-0.5/wgh[dex](vofs[vi],c-1);
                  cout << " comp " << res[dex](vofs[vi],c-1) << " in class " << 0.5/wgh[dex](vofs[vi],c-1) ;
		  if( fabs(dif)>1e-6*fabs(res[dex](vofs[vi],c-1)) )
		    cout << "   <------- " ;
		  cout << endl;
		  //                  ebis[dex].volFrac(vofs[vi]) ebis[dex].normal(vofs[vi]) ebis[dex].bndryArea(vofs[vi]) ebis[dex].bndryArea(vofs[vi],
	       }
	    }
	 }
      }
   }
   else if( testcase == -4 || testcase == -5 )
   {
      cout << "------------------------------------------------------------\n";
      cout << "Test -4, Helmholtz equation.\n" <<
	      "Diagonal terms of operator matrix." << endl;
      cout << "------------------------------------------------------------\n";

      int l = 0;

      Real relpar = 1.0;
      if( relax_type == 0 )
	 relpar = 0.5;
      
      hamrop = hamrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );

      EBCellFactory cfactory( ebises[l] );
      LevelData<EBCellFAB>   uus( dbls[l], 1, IntVect::Zero, cfactory );
      LevelData<EBCellFAB>   res( dbls[l], 1, IntVect::Zero, cfactory );
      LevelData<EBCellFAB>   wgh( dbls[l], 1, IntVect::Zero, cfactory );

      const EBISLayout& ebis = ebises[l];
      for( DataIterator li=dbls[l].dataIterator() ; li.ok() ; ++li )
      {
         DataIndex dex(li());
         for( BoxIterator b(ebis[dex].getRegion()); b.ok(); ++b )
         {
            if( !ebis[dex].isCovered(b()) )
	    {
	       Vector<VolIndex> vofs = ebis[dex].getVoFs(b());
               for( int vi=0 ; vi < vofs.size() ; vi++ )
  	       {
                  res[dex].setVal(0.0);
		  uus[dex].setVal(0.0);
		  uus[dex](vofs[vi],0) = 1;
		  hamrop->applyOp( res, uus, NULL, true, true, NULL );
                  cout << "( " << b()[0] <<" , "<< b()[1] << " ) ";
                  if( ebis[dex].isRegular(b()) )
                    cout << "r" ;
		  else
		    cout << "i" ;
                  Real dif = res[dex](vofs[vi],0)-relpar/hamrop->get_diagonal_weight(dex,vofs[vi]);
                  cout << " comp " << res[dex](vofs[vi],0) << " in class " << relpar/hamrop->get_diagonal_weight(dex,vofs[vi]);
		  if( fabs(dif)>1e-6*fabs(res[dex](vofs[vi],0)) )
		    cout << "   <------- " ;
		  cout << endl;
	       }
	    }
	 }
      }
   }
   else if( testcase == -6 )
   {
      ofstream filut("shape.dat");
      int l=0;
      filut << dx[l][0] << "  " << dx[l][1] << " " << 0.0 << " " << 0.0 << " " << 0.0 << " " << 0.0 << " " << 0.0 << " " << 0.0 << " " << 0.0 << " " << 0.0 << " " << 0.0 << " " << 0.0 << endl;
      get_testdata( *u[l], ebises[l], dx[l], x0, 6 );
      const LevelData<EBCellFAB>& ub = *u[l];
      for( DataIterator li=dbls[l].dataIterator() ; li.ok() ; ++li )
      {
         DataIndex dex(li());
         for( BoxIterator b(ebises[l][dex].getRegion()); b.ok(); ++b )
         {
	    Vector<VolIndex> vofar = ebises[l][dex].getVoFs(b());
	    for( int iv=0 ; iv < vofar.size() ; iv++ )
	    {
	       cout << "---------------------------------------" << endl;
	       cout << b() << " u at vof is " << ub[dex](vofar[iv],0) << endl;
               cout << "Cell index is " << vofar[iv].cellIndex() << endl;

               Vector<VolIndex> vn = ebises[l][dex].getVoFs(vofar[iv],0,Side::Lo,1);
               for( int i=0 ; i<vn.size() ; i++ )
		  cout << "left neigbor is  " << ub[dex](vn[i],0) << endl;

               vn = ebises[l][dex].getVoFs(vofar[iv],0,Side::Hi,1);
               for( int i=0 ; i<vn.size() ; i++ )
		  cout << "Right neigbor is " << ub[dex](vn[i],0) << endl;

               vn = ebises[l][dex].getVoFs(vofar[iv],1,Side::Lo,1);
               for( int i=0 ; i<vn.size() ; i++ )
		  cout << "Down neigbor is  " << ub[dex](vn[i],0) << endl;

               vn = ebises[l][dex].getVoFs(vofar[iv],1,Side::Hi,1);
               for( int i=0 ; i<vn.size() ; i++ )
		  cout << "Up neigbor is    " << ub[dex](vn[i],0) << endl;

               cout << "Normal is " << ebises[l][dex].normal(vofar[iv]) << endl; 
               cout << "Cell is " ;
               if( ebises[l][dex].isCovered(vofar[iv].gridIndex()) )
		  cout << "c" ;
	       else if( ebises[l][dex].isIrregular(vofar[iv].gridIndex()) )
		  cout << "i";
	       else
		  cout << "r";
	       cout << endl;
	       cout << "centroid is" << ebises[l][dex].centroid(vofar[iv]) << endl;
              cout << "bndry centroid is  " << ebises[l][dex].bndryCentroid(vofar[iv]) << endl;
              if( ebises[l][dex].volFrac(vofar[iv]) < 1 )
              out_shape( ebises[l][dex].normal(vofar[iv]),
			 ebises[l][dex].bndryCentroid(vofar[iv]),
			 vofar[iv].gridIndex(),
			 ebises[l][dex].centroid(vofar[iv]),
			 dx[0],
			 filut );
	      cout << "volume fraction is " << ebises[l][dex].volFrac(vofar[iv]) << endl;
               Vector<FaceIndex> faces = ebises[l][dex].getFaces(vofar[iv],1,Side::Hi);
               for( int i=0 ; i<faces.size() ; i++ )
	       {
 	           VolIndex vh=faces[i].getVoF(Side::Hi);
	           VolIndex vl=faces[i].getVoF(Side::Lo);

                   if( vh.cellIndex() >-1 && vl.cellIndex() > -1 )
		       cout << "Face between " << ub[dex](vh,0) << " and " << ub[dex](vl,0);
                   else if( vh.cellIndex() > -1 )
		       cout << "low domain bndry at " << ub[dex](vh,0);
		   else
		       cout << "high domain bndry at " << ub[dex](vl,0);
                   cout << "centroid " << ebises[l][dex].centroid (faces[i]) << endl;
	       }         
               Vector<VolIndex> allMonotoneVoFs;
               int radius = 1;
               IntVect timesMoved = IntVect::Zero;
               IntVect pathSign   = IntVect::Zero;

               EBArith::getAllVoFsInMonotonePath(allMonotoneVoFs,timesMoved,pathSign,
                                    vofar[iv],ebises[l][dex],radius);
               cout << "Monotone vofs of radius "<< radius << endl;
         
               for( int i=0 ; i< allMonotoneVoFs.size() ; i++ )
		  cout <<  "         " << ub[dex](allMonotoneVoFs[i],0) << endl;

	    }
      	 }
         cout << "I-Fluxes: " << endl;
         IntVectSet all_points(ebises[l][dex].getRegion());
         for( FaceIterator it(all_points,ebises[l][dex].getEBGraph(), 0,
			      FaceStop::SurroundingNoBoundary); it.ok(); ++it)
	 {
	    VolIndex vh=it().getVoF(Side::Hi);
	    VolIndex vl=it().getVoF(Side::Lo);
            cout << " side between " << ub[dex](vh,0) << " and " << ub[dex](vl,0) ;
            cout << "centroid " << ebises[l][dex].centroid(it()) << endl;
	 }
      }
   }
   else if( testcase == 10 )
   {
// Evaluate operator on known function, print out trunc errors.
      cout << "------------------------------------------------------------\n";
      cout << "Test 10, Helmholtz equation with solution sin(x)*cos(y).\n" <<
	      "Testing truncation error in cylindrical coordinates." << endl;
      cout << "------------------------------------------------------------\n";


// Get test data
      for( int l= 0 ; l < nlevels ; l++ )
         get_testdata( *u[l], ebises[l], dx[l], x0, 5 );
	//         get_testdata( *u[l], ebises[l], dx[l], x0, 7 );

// Apply operator
      for( int l = 0; l < nlevels ; l++ )
      {
         hamrop = hamrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         if( l == 0 )
            hamrop->applyOp( *res[l], *u[l], NULL, false, false, NULL );
         else
            hamrop->applyOp( *res[l], *u[l], u[l-1], false, false, NULL );
         if( l < nlevels-1 )
	    hamrop->reflux( *res[l], *u[l+1], *u[l] ); 
// Get exact residual
         get_exactres( *rhs[l], ebises[l], dx[l], x0, 9, 0, k2, my, la );

// volume divided or not
	 if( resmulvol )
	    multiply_volume( *rhs[l], ebises[l] );
	 else
            divide_volume( *res[l], ebises[l] );
      }
      output_level_array( res, ebises, dx, x0, "resl.bin");
      output_level_array( rhs, ebises, dx, x0, "exl.bin" );

      for( int l=0 ; l < nlevels ; l++ )
      {
         EBLevelDataOps::scale(*rhs[l],-1.0);
         EBLevelDataOps::sum( *err[l], *rhs[l], *res[l] );
         hamrop = hamrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Error norm L2:   " << hamrop->norm(*err[l],2);
         cout << "Error norm Linf: " << hamrop->norm(*err[l],0) << endl;
      }
   }
   else if( testcase == 11 )
   {

// Solve elliptic problem for a known forcing
     cout << "-------------------------------------------------------------\n";
     cout << "Test 11, Helmholtz equation with solution sin(x)*cos(y)." <<
	     "Solving the problem in cylindrical coordinates." << endl;
     cout << "-------------------------------------------------------------\n";

     for( int l = 0; l < nlevels ; l++ )
     {
        get_exactres( *rhs[l], ebises[l], dx[l], x0, 9, 1, k2, my, la );
// Initial guess
        EBLevelDataOps::setToZero(*u[l]);
     }

// Set up MG solver
      if( sol == 1 )
      {
         AMRMultiGrid<LevelData<EBCellFAB> > mgsolver;
         BiCGStabSolver<LevelData<EBCellFAB> > bottomSolver;      
	 //         EBSimpleSolver bottomSolver;      
         int maxlevels = nlevels;
         mgsolver.define( doms[0], *hamrop_factory, &bottomSolver, maxlevels );

         int l_max = nlevels-1, l_base = 0;
         mgsolver.solve(  u, rhs, l_max, l_base );
      }
      else if( sol == 2 )
      {
	 for( int l = 0 ; l < nlevels ; l++ )
	 {
            hamrop = hamrop_factory->createOperator( dbls[l], ebises[l], doms[l],
						     dx[l], l );
	    hamrop->set_inhomogeneous_relax();
	    hamrop->relax( *u[l], *rhs[l], nit );
	 }
      }

      output_level_array( u, ebises, dx, x0, "ul.bin" );

// Get exact solution and compare
      for( int l= 0 ; l<nlevels ; l++ )
         get_testdata( *res[l], ebises[l], dx[l], x0, 5 );

      output_level_array( res, ebises, dx, x0, "exl.bin" );

      for( int l = 0 ; l < nlevels ; l++ )
      {
         EBLevelDataOps::scale(*res[l],-1.0);
         EBLevelDataOps::sum( *err[l], *u[l], *res[l] );
         hamrop = hamrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Error norm L2:   " << hamrop->norm(*err[l],2);
         cout << "Error norm Linf: " << hamrop->norm(*err[l],0) << endl;
      }
   }
   else if( testcase == 12 )
   {
// Test energy equation work term
     cout << "-------------------------------------------------------------\n";
     cout << "Test 12, Navier-Stokes energy equation work term" <<
	     "Truncation error in Cartesizn coordinates." << endl;
     cout << "-------------------------------------------------------------\n";

// Get test data
      for( int l = 0; l < nlevels ; l++ )
         get_testdata( *u[l], ebises[l], dx[l], x0, 2 );

// Apply operator
      for( int l = 0 ; l < nlevels ; l++ )
      {
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );

         if( l == 0 &&  nlevels>1 )
	    visc_amrop->AMRResidual_eeq( *res[l], u[l+1], *u[l], NULL, false );
	 else if( l== 0 && nlevels == 1 )
	    visc_amrop->AMRResidual_eeq( *res[l], NULL, *u[l], NULL, false );
	 else if( l < nlevels-1 )
	    visc_amrop->AMRResidual_eeq( *res[l], u[l+1], *u[l], u[l-1], false );
         else
	    visc_amrop->AMRResidual_eeq( *res[l], NULL, *u[l], u[l-1], false );

         get_exactres( *rhs[l], ebises[l], dx[l], x0, 7, 0, cof, my, la );

	 if( resmulvol )
	    multiply_volume( *rhs[l], ebises[l] );
	 else
            divide_volume( *res[l], ebises[l] );
      }
// Get exact residual and compare
      output_level_array( res, ebises, dx, x0, "resl.bin");
      output_level_array( rhs, ebises, dx, x0, "exl.bin" );

      for( int l = 0 ; l < nlevels ; l++ )
      {
         EBLevelDataOps::scale(*rhs[l],-1.0);
         EBLevelDataOps::sum( *err[l], *rhs[l], *res[l] );
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Error norm L2: " << visc_amrop->norm(*err[l],2);
         cout << "Error norm Linf: " << visc_amrop->norm(*err[l],0) << endl;
      }
   }
   else if( testcase == 13 )
   {
// Test energy equation work term
     cout << "-------------------------------------------------------------\n";
     cout << "Test 13, Navier-Stokes energy equation work term" <<
	     "Truncation error in Cylindrical coordinates." << endl;
     cout << "-------------------------------------------------------------\n";

// Get test data
      for( int l = 0; l < nlevels ; l++ )
         get_testdata( *u[l], ebises[l], dx[l], x0, 4 );

// Apply operator
      for( int l = 0 ; l < nlevels ; l++ )
      {
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );

         if( l == 0 &&  nlevels>1 )
	    visc_amrop->AMRResidual_eeq( *res[l], u[l+1], *u[l], NULL, false );
	 else if( l== 0 && nlevels == 1 )
	    visc_amrop->AMRResidual_eeq( *res[l], NULL, *u[l], NULL, false );
	 else if( l < nlevels-1 )
	    visc_amrop->AMRResidual_eeq( *res[l], u[l-1], *u[l], u[l+1], false );
         else
	    visc_amrop->AMRResidual_eeq( *res[l], u[l-1], *u[l], NULL, false );

         get_exactres( *rhs[l], ebises[l], dx[l], x0, 10, 0, cof, my, la );
	 if( resmulvol )
	    multiply_volume( *rhs[l], ebises[l] );
	 else
            divide_volume( *res[l], ebises[l] );
      }
// Get exact residual and compare
      output_level_array( res, ebises, dx, x0, "resl.bin");
      output_level_array( rhs, ebises, dx, x0, "exl.bin" );

      for( int l = 0 ; l < nlevels ; l++ )
      {
         EBLevelDataOps::scale(*rhs[l],-1.0);
         EBLevelDataOps::sum( *err[l], *rhs[l], *res[l] );
         visc_amrop = visc_amrop_factory->createOperator( dbls[l], ebises[l], doms[l], dx[l], l );
         cout << "Error norm L2: " << visc_amrop->norm(*err[l],2);
         cout << "Error norm Linf: " << visc_amrop->norm(*err[l],0) << endl;
      }
   }
}
