/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _EBAMRNSFACTORY_H_
#define _EBAMRNSFACTORY_H_

#include "EBAMRNS.H"
#include "AMRLevelFactory.H"
#include "EBPatchGodunovFactory.H"

///
/**
 */
class EBAMRNSFactory : public AMRLevelFactory
{
public:
  ///
  AMRLevel* new_amrlevel() const;

  ///
  EBAMRNSFactory(     const Real&                        a_initialDtMultiplier,
                      const Real&                        a_cfl,
                      const int &                        a_redistRad,
                      const Real&                        a_domainLength,
                      const Real&                        a_refineThresh,
                      const int &                        a_tagBufferSize,
                      const int &                        a_verbosity,
                      const bool&                        a_useMassRedist,
                      const bool&                        a_doSmushing,
                      const bool&                        a_doRZCoords,
                      const bool&                        a_hasSourceTerm,
                      const EBPatchGodunovFactory* const a5_patchGodunov,
		      const Real&                        a_gamma  = 1.4, 
		      const Real&                        a_rom    = 277.14, 
			       const Real& a_mu=0.1,
			       const Real& a_lambda=0.1,
			       const Real& a_kappa=0.1,
                      const int& a_force = 0,
                      bool                               a_tagAll = false );


  ///
  virtual ~EBAMRNSFactory();

protected:
  bool                   m_tagAll;
  Real                   m_initialDtMultiplier;
  Real                   m_cfl;
  int                    m_redistRad;
  Real                   m_domainLength;
  Real                   m_refineThresh;
  Real                   m_gamma;
  Real                   m_rom;
  Real                   m_mu;
  Real                   m_lambda;
  Real                   m_kappa;
  int                    m_force;
  int                    m_tagBufferSize;
  int                    m_verbosity;
  bool                   m_useMassRedist;
  bool                   m_doSmushing;
  bool                   m_doRZCoords;
  bool                   m_hasSourceTerm;
  const EBPatchGodunovFactory* m_patchGodunov;


private:
  EBAMRNSFactory()
  {
    MayDay::Error("invalid operator");
  }
  //disallowed for all the usual reasons
  void operator=(const EBAMRNSFactory& a_input)
  {
    MayDay::Error("invalid operator");
  }
  EBAMRNSFactory(const EBAMRNSFactory& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
