/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include "EBAMRNS.H"
#include "EBAMRNSFactory.H"


//-----------------------------------------------------------------------
AMRLevel* EBAMRNSFactory::new_amrlevel() const
{
  EBAMRNS* amrg_ptr = new EBAMRNS();

  amrg_ptr->CFL(m_cfl);
  amrg_ptr->patchGodunov(m_patchGodunov);
  amrg_ptr->refinementThreshold(m_refineThresh);
  amrg_ptr->tagBufferSize(m_tagBufferSize);
  amrg_ptr->initialDtMultiplier(m_initialDtMultiplier);
  amrg_ptr->domainLength(m_domainLength);
  amrg_ptr->redistRadius(m_redistRad);
  amrg_ptr->verbosity(m_verbosity);
  amrg_ptr->useMassRedistribution(m_useMassRedist);
  amrg_ptr->doRZCoords(m_doRZCoords);
  amrg_ptr->hasSourceTerm(m_hasSourceTerm);
  amrg_ptr->doSmushing(m_doSmushing);
  amrg_ptr->tagAll(m_tagAll);
  amrg_ptr->set_gamma(m_gamma);
  amrg_ptr->set_rom(m_rom);
  amrg_ptr->set_viscosities( m_mu, m_lambda, m_kappa );
  amrg_ptr->set_forcing( m_force );
  return (static_cast <AMRLevel*> (amrg_ptr));
}

//-----------------------------------------------------------------------
EBAMRNSFactory::~EBAMRNSFactory()
{

}

//-----------------------------------------------------------------------
EBAMRNSFactory::EBAMRNSFactory(const Real&        a_initialDtMultiplier,
			       const Real&                   a_cfl,
                    const int &                   a_redistRad,
                    const Real&                   a_domainLength,
                    const Real&                   a_refineThresh,
                    const int &                   a_tagBufferSize,
                    const int &                   a_verbosity,
                    const bool&                   a_useMassRedist,
                    const bool&                   a_doSmushing,
                    const bool&                   a_doRZCoords,
                    const bool&                   a_hasSourceTerm,
			       const EBPatchGodunovFactory* const a_patchGodunov,
                    const Real&                   a_gamma,
		    const Real&                   a_rom,
		    const Real&                   a_mu,
		    const Real&                   a_lambda,
		    const Real&                   a_kappa,
	            const int&                    a_force,
		    bool                          a_tagAll )
{
  m_tagAll = a_tagAll;

  m_initialDtMultiplier = a_initialDtMultiplier;
  m_cfl                 = a_cfl;
  m_useMassRedist       = a_useMassRedist;
  m_redistRad           = a_redistRad;
  m_domainLength        = a_domainLength;
  m_refineThresh        = a_refineThresh;
  m_tagBufferSize       = a_tagBufferSize;
  m_verbosity           = a_verbosity;
  m_doSmushing          = a_doSmushing;
  m_doRZCoords          = a_doRZCoords;
  m_hasSourceTerm       = a_hasSourceTerm;
  m_patchGodunov        = a_patchGodunov;
  m_gamma               = a_gamma;
  m_rom                 = a_rom;
  m_mu     = a_mu;
  m_lambda = a_lambda;
  m_kappa  = a_kappa;
  m_force  = a_force;
}

