/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _HELMHOLTZEBRZBC_H_
#define _HELMHOLTZEBRZBC_H_

#include "RefCountedPtr.H"
#include "BaseBCVectValue.H"
#include "EBCellFAB.H"
#include "EBISLayout.H"

///
/**
  x
 */
class HelmholtzEBBC
{
public:
  /// Constructor.
  /**
   */
  HelmholtzEBBC();

  /// Destructor.
  /**
   */
  virtual ~HelmholtzEBBC();

  /// Order of accuracy in boundary normal derivative approximation.
  /**
   */
  void set_order( int a_order );

  /// Give a constant value on the boundary.
  /**
   */
  void set_value( Vector<Real> a_value );

  /// Give a boundary value that varies on the boundary.
  /**
   */
  void set_function( RefCountedPtr<BaseBCVectValue> a_func );

  /// BC for axisymmetric equations
  /**
   */
  void set_rz( bool a_rz );

  /// Compute the boundary flux.
  /**
   */
  virtual void get_ebflux(Vector<Real>&               a_flux,
                          const VolIndex&             a_vof,
                          const LevelData<EBCellFAB>& a_phi,
                          const DataIndex&            a_dit,
                          const RealVect&             a_probLo,
                          const RealVect&             a_dx,
                          bool                        a_use_homogeneous,
                          Real                        a_time ) = 0;

protected:
  bool m_use_value;
  bool m_use_function;
  bool m_rz;
  int m_order;

  Vector<Real> m_value;
  RefCountedPtr<BaseBCVectValue> m_func;

};

/// Class that can generate a HelmholtzEBRZBC object.
/**
 */
class HelmholtzEBBCFactory
{
public:
  /// Constructor.
  /**
   */
  HelmholtzEBBCFactory();

  /// Destructor.
  /**
   */
  virtual ~HelmholtzEBBCFactory();

  /// Order of accuracy in boundary normal derivative approximation.
  /**
   */
  virtual void set_order( int a_order );

  /// Give a constant value on the boundary.
  /**
   */
  void set_value( Vector<Real> a_value );

  /// Give a boundary value that varies on the boundary.
  /**
   */
  void set_function( RefCountedPtr<BaseBCVectValue> a_func );


  /// BC for axisymmetric equations
  /**
   */
  void set_rz( bool a_rz );

  /// Generate a HelmholtzEBBC object.
  /**
   */
  virtual HelmholtzEBBC* create( const ProblemDomain& a_domain,
				 const EBISLayout&    a_layout,
				 const RealVect&      a_dx ) = 0;

protected:
  bool m_use_value;
  bool m_use_function;
  int m_order;
  bool m_rz;
  Vector<Real> m_value;
  RefCountedPtr<BaseBCVectValue> m_func;
};


#endif
