/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _AMRLEVELMHDFACTORY_H_
#define _AMRLEVELMHDFACTORY_H_

#include "AMRLevelFactory.H"

#include "AMRLevelMHD.H"

/// factory object for creating AMRLevelMHD objects
/**
 */
class AMRLevelMHDFactory : public AMRLevelFactory
{
public:
  /// Null constructor
  /**
   */
  AMRLevelMHDFactory();

  /// Destructor
  /**
   */
  virtual ~AMRLevelMHDFactory();

  /// Define the parameters of the factory
  /**
   */
  virtual void define(const Real&                 a_cfl,
                      const Real&                 a_domainLength,
                      const int&                  a_verbosity,
                      const Real&                 a_refineThresh,
                      const int&                  a_tagBufferSize,
                      const Real&                 a_initialDtMultiplier,
                      const GodunovPhysics* const a_godunovPhysics,
                      const int&                  a_normalPredOrder,
                      const bool&                 a_useFourthOrderSlopes,
                      const bool&                 a_usePrimLimiting,
                      const bool&                 a_useCharLimiting,
                      const bool&                 a_useFlattening,
                      const bool&                 a_useArtificialViscosity,
                      const Real&                 a_artificialViscosity,
                      const bool&                 a_doFilterBField);

  /// Create a new AMRLevel (for MHD)
  /**
   */
  virtual AMRLevel* new_amrlevel() const;

  /// Check that this object has been defined
  /**
   */
  bool isDefined() const;


protected:
  // CFL number
  Real m_cfl;

  // Physical dimension of the longest side of the domain
  Real m_domainLength;

  // Verbosity of the object
  int m_verbosity;

  // Refinement threshold for gradient
  Real m_refineThresh;

  // Tag buffer size
  int m_tagBufferSize;

  // Initial dt multiplier
  Real m_initialDtMultiplier;

  // Pointer to the object that supplies the physics needed by the integrator
  // (used as a factory here)
  const GodunovPhysics* m_godunovPhysics;

  // Order of the normal predictor (1 -> PLM, 2 -> PPM)
  int m_normalPredOrder;

  // Use 4th other slope computations (otherwise, use 2nd order)
  bool m_useFourthOrderSlopes;

  // Do slope limiting in the primitive or characteristic variables, respect.
  bool m_usePrimLimiting;
  bool m_useCharLimiting;

  // Do slope flattening - MUST BE USING 4th order slopes
  bool m_useFlattening;

  // Apply artificial viscosity of a set value
  bool m_useArtificialViscosity;
  Real m_artificialViscosity;

  // Apply filter to magnetic field
  bool m_doFilterBField;
  // Has this object been defined
  bool m_isDefined;


private:
  // Disallowed for all the usual reasons
  void operator=(const AMRLevelMHDFactory& a_input)
  {
    MayDay::Error("invalid operator");
  }

  // Disallowed for all the usual reasons
  AMRLevelMHDFactory(const AMRLevelMHDFactory& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
