# -*- Mode: Makefile; Modified: "Tue 06 Sep 2005 10:48:04 by dbs"; -*-

### This is a compiler-specific definitions file for GNU compilers.

## It sets values for the defaults variables ($def*), which will be used
## if the calling makefile doesn't set the corresponding variables.
## The corresponding variables are the ones without the "def" prefix.

## Defaults variables:
##  defcppdbgflags = C-preprocessor options for both C++ and Fortran code when DEBUG=TRUE
##  defcppoptflags = C-preprocessor options for both C++ and Fortran code when DEBUG=FALSE
##  defcxxcppflags = C-preprocessor options for only C++ code
##  defcxxdbgflags : for C++ compiling and linking when DEBUG=TRUE
##  defcxxoptflags : for C++ compiling and linking when DEBUG=FALSE
##  defcxxprofflags: for C++ compiling and linking when PROFILE=TRUE
##  deffcppflags   = C-preprocessor options for only Fortran code
##  deffdbgflag    : for Fortran compiling when DEBUG=TRUE
##  deffoptflag    : for Fortran compiling when DEBUG=FALSE
##  deffprofflags  : for Fortran compiling when PROFILE=TRUE
##  defflibflags   : library options needed to link the Fortran code
##  deflddbgflags  = linker options when DEBUG=TRUE
##  defldoptflags  = linker options when DEBUG=FALSE
##  defldprofflags = linker options when PROFILE=TRUE
##  cppcallsfort   = preprocessor macro as used in FORT_PROTO.H to specify
##                    how the C++ code should construct names for calling Fortran
##
## Special object models:
##  OBJMODEL = 32   Compile/link with 32bit pointers
##  OBJMODEL = 64   Compile/link with 64bit pointers
##
##  Note: 64bit only works on AMD64 computers with gcc 3.3 or later.  <dbs,Mar04>
##
## Notes:
##  GNU C/C++ compilers set the preprocessor macros: (use -E -dM options to display)
##    __GNUC__       (int) defined for gcc or g++, == major version number
##    __GNUG__       (int) defined for g++, == major version number
##    __GNUC_MINOR__ (int) defined for gcc or g++, == minor version number

makefiles+=compiler/Make.defs.GNU

cxxname := $(notdir $(firstword $(CXX)))
fname   := $(notdir $(firstword $(FC)))

# special case for g++ version3 installed on a system where version2 is the default
ifeq ($(cxxname),g++3)
  cxxname := g++
endif
ifeq ($(fname),g773)
  fname := g77
endif

################################################################

ifeq ($(cxxname),g++)
  _gppversion := $(subst ., ,$(shell $(CXX) -dumpversion))
  _gppmajorver:= $(word 1,$(_gppversion))
  _gppminorver:= $(word 2,$(_gppversion))

  # `-ftemplate-depth' is needed to keep mangled function names from
  #  getting too long for some linkers to handle.
  # HDF5 uses `long long', so disable warnings about it.
  # `-Wno-sign-compare' turns off warnings about comparing signed and unsigned int -
  #   this is a meaningful warning, but it appears way too often in our code.

  _cxxbaseflags := -Wno-long-long -Wno-sign-compare -Wno-deprecated -ftemplate-depth-25

  # g++ v3.4.0 started objecting to some uses of HOFFSET()
  # TJL - Changed moving from csh to sh...
  # ifeq (00,$(shell test $(_gppmajorver) -ge 3 ; echo $$status)$(shell test $(_gppminorver) -ge 4 ; echo $$status))
  ifeq (00,$(shell test $(_gppmajorver) -ge 3 ; echo $$?)$(shell test $(_gppminorver) -ge 4 ; echo $$?))
    _cxxbaseflags += -Wno-invalid-offsetof
  endif

  # NOTE: -malign-double causes MPI programs that use pout() to
  # segfault in ~ostream(), so dont use it
  defcxxoptflags := -O2 -funroll-loops $(_cxxbaseflags)
  defcxxdbgflags := -g -pedantic -Wall $(_cxxbaseflags)
  # gcc v4 doesn't understand this yet
  ifeq (0,$(shell test $(_gppmajorver) -lt 4 ; echo $$?))
    defcxxdbgflags += -fbounds-checking
  endif
  # add extra symbol info to debug output for gdb
  ifeq ($(system),OSF1)
    # -gstabs breaks on gcc 3.3 and osf1 5.1 (on halem)
    # its not clear which is the cause, so for now I'll assume gcc is the problem
    ifeq ($(_gppmajorver)$(_gppminorver),33)
      defcxxdbgflags += -ggdb
    else
      defcxxdbgflags += -gstabs
    endif
  else
    defcxxdbgflags += -ggdb
  endif

  # non-default object models are meaningful ONLY on some machines
  ifeq ($(OBJMODEL),64)
    ifeq ($(arch),x86_64)
      # works on an AMD64 machine that defaults to 32bit
      # may also work on a Power{PC,4,5} but I'm not positive
      defcxxdbgflags += -m64
      defcxxoptflags += -m64
    endif
    # on IBM Power (like seaborg)
    ### WARNING: I haven't tested this <dbs,5Mar04> !!!
    ifeq ($(system),AIX)
      defcxxdbgflags += -maix64
      defcxxoptflags += -maix64
    endif
  endif
  ifeq ($(OBJMODEL),32)
    ifeq ($(arch),x86_64)
      # on AMD64 machines with default 64bit, you have to ask
      # for an emulation mode to get 32 bits
      defcxxdbgflags += -m32
      defcxxoptflags += -m32
      deflddbgflags = -Wl,-melf_i386
      defldoptflags = -Wl,-melf_i386
    endif
    # on IBM Power with default 64bit mode
    ### WARNING: I haven't tested this <dbs,5Mar04> !!!
    ifeq ($(system),AIX)
      defcxxdbgflags += -maix32
      defcxxoptflags += -maix32
    endif
  endif

  # the preprocessor in v3 or later strips "//" by default,
  # but this is a valid Fortran operator, so force it to leave comments
  ifeq (0,$(shell test $(_gppmajorver) -ge 3 ; echo $$?))
    CPP += -C
  endif
endif

################################################################

ifeq ($(fname),g77)
  _g77version := $(subst ., ,$(shell $(FC) -dumpspecs | awk '/^\*version:/{getline;print $$1}'))
  _g77majorver:= $(word 1,$(_g77version))
  _g77minorver:= $(word 2,$(_g77version))

  deffoptflags = -O2 -funroll-loops -fno-second-underscore
  # if machine has i686 in the uname -a, then add the -malign-double opt flag
  ifeq ($(findstring i686,$(unamea)),i686)
    deffoptflags += -malign-double
  endif

  deffdbgflags = -g -Wimplicit -Wall -fno-second-underscore -ffortran-bounds-check

  # add extra symbol info to debug output for gdb
  ifeq ($(system),OSF1)
    # -gstabs breaks on gcc 3.3 and osf1 5.1 (on halem)
    # its not clear which is the cause, so for now I'll assume gcc is the problem
    ifeq ($(_g77majorver)$(_g77minorver),33)
      deffdbgflags += -ggdb
    else
      deffdbgflags += -gstabs
    endif
  else
    deffdbgflags += -ggdb
  endif

  # non-default object models are meaningful ONLY on some machines
  ifeq ($(OBJMODEL),64)
    ifeq ($(arch),x86_64)
      # works on an AMD64 machine that defaults to 32bit
      # may also work on a Power{PC,4,5} but I'm not positive
      deffdbgflags += -m64
      deffoptflags += -m64
    endif
    # on IBM Power (like seaborg)
    ### WARNING: I haven't tested this <dbs,5Mar04> !!!
    ifeq ($(system),AIX)
      deffdbgflags += -maix64
      deffoptflags += -maix64
    endif
  endif
  ifeq ($(OBJMODEL),32)
    ifeq ($(arch),x86_64)
      # on AMD64 machines with default 64bit, you have to ask
      # for an emulation mode to get 32 bits
      deffdbgflags += -m32
      deffoptflags += -m32
    endif
    # on IBM Power with default 64bit mode
    ### WARNING: I haven't tested this <dbs,5Mar04> !!!
    ifeq ($(system),AIX)
      deffdbgflags += -maix32
      deffoptflags += -maix32
    endif
  endif

  # libraries needed by g77
  # g++ v2 doesn't have libfrtbegin but g77 v3 needs it
  _libg2c_dir := $(dir $(shell $(FC) -print-file-name=libg2c.a))
  ifeq ($(_libg2c_dir),)
    $(error cannot find libg2c.a using $(FC) -print-file-name=libg2c.a)
  endif
  _libfrt_dir := $(dir $(shell $(FC) -print-file-name=libfrtbegin.a))
  # set library directory (if not using g++) and library names; use libfrt only if it exists
  ifeq ($(cxxname),g++)
    defflibflags := -lg2c
  else
    defflibflags := -L$(_libg2c_dir) -lg2c
  endif
  ifneq ($(_libfrt_dir),./)  #[NOTE: $dir returns ./ when there is no directory part]
    # in the unlikely event libg2c and libfrtbegin are in different directories
    ifneq ($(_libfrt_dir),$(_libg2c_dir))
      defflibflags += -L$(_libfrt_dir)
    endif
    defflibflags += -lfrtbegin
  endif
  defflibflags += -lm
endif

################################################################

# GNU Fortran95 compiler in GCC version 4
ifeq ($(fname),gfortran)
  _g77version := $(subst ., ,$(shell $(FC) -dumpspecs | awk '/^\*version:/{getline;print $$1}'))
  _g77majorver:= $(word 1,$(_g77version))
  _g77minorver:= $(word 2,$(_g77version))

  deffoptflags = -O2 -funroll-loops -Wno-line-truncation -fno-second-underscore
  deffdbgflags = -g -ggdb -Wall -W  -Wno-line-truncation -fno-second-underscore#  #-fbounds-checking not supported in v4.0.0
  defflibflags  = -lgfortranbegin -lgfortran -lm
endif

################################################################

# Special link flags for DEC/Compaq/HP Tru64/OSF1
ifeq ($(system),OSF1)
  deflddbgflags = -Wl,-S
  defldoptflags = -Wl,-S
endif
