/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// DFMartin, Sun, May 5, 2002

#ifndef _BASEBOTTOMSMOOTHER_H_
#define _BASEBOTTOMSMOOTHER_H_

#include <iostream>
#include <cmath>
#include "SPACE.H"
#include <cstdlib>

#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "ProblemDomain.H"
#include "LevelOp.H"

/// Pure base class to encapsulate bottom smooth for LevelOp-derived objects.
/** The BaseBottomSmoother class is a virtual base class for implementing
    different bottomSmoother algorithms for LevelOp-derived classes. It
    takes a LevelOp pointer as an argument during its definition, in order
    to access the problem-dependent (equation-dependent) functionality of
    the specific LevelOp derivative class.
*/
class BaseBottomSmoother
{
public:
  /// Null constructor
  BaseBottomSmoother()
  {
  }

  /// Destructor
  virtual ~BaseBottomSmoother()
  {
  }

  /// This gets around the "no virtual constructor" rule
  virtual BaseBottomSmoother* new_bottomSmoother() const = 0;

  /// Do smoothing
  virtual void doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                              const LevelData<FArrayBox>& a_rhs,
                              LevelOp*                    a_levelopPtr) = 0;

  /// sets metric to judge convergence (normally norm(rhs) of outer problem
  /** To make the iterative solver convergence more consistent, the solver
      will consider itself "solved" if the residual is reduced to
      epsilon*metric, where epsilon is the solver tolerance (set by
      setSolverTol).  A good starting point for the metric is the norm of
      the RHS of the original problem (i.e. the rhs sent in to LevelSolver
      or AMRSolver).  Default value is 1.0.
   */
  virtual void setConvergenceMetric(Real a_metric, int a_comp=0) = 0;

  /// set solver tolerance
  virtual void setSolverTol(Real a_tol) = 0;

};

#endif
