/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <iostream>
using std::cout;
using std::cin;
using std::cerr;
using std::endl;

#include "Box.H"
#include "Vector.H"
#include "DisjointBoxLayout.H"
#include "LayoutIterator.H"
#include "parstream.H"

#include "AMRLevel.H"

int AMRLevel::s_verbosity = 0;

bool AMRLevel::isDefined() const
{
  return m_isDefined;
}

AMRLevel::~AMRLevel()
{
}

AMRLevel::AMRLevel()
{
  m_coarser_level_ptr = NULL;
  m_finer_level_ptr = NULL;
  m_isDefined = false;
  m_level = 0;
  m_time = 0;
  m_dt = 0;
  m_initial_dt_multiplier = 0.1;
}

void AMRLevel::define(AMRLevel*  a_coarser_level_ptr,
                      const Box& a_problem_domain,
                      int        a_level,
                      int        a_ref_ratio)
{
  ProblemDomain physDomain(a_problem_domain);

  define(a_coarser_level_ptr, physDomain, a_level, a_ref_ratio);
}

void AMRLevel::define(AMRLevel*            a_coarser_level_ptr,
                      const ProblemDomain& a_problem_domain,
                      int                  a_level,
                      int                  a_ref_ratio)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevel::define" << endl;
  }

  m_coarser_level_ptr = a_coarser_level_ptr;
  m_problem_domain = a_problem_domain;
  m_level = a_level;
  m_ref_ratio = a_ref_ratio;
  m_finer_level_ptr = NULL;
  m_isDefined = true;
}

void AMRLevel::finerLevelPtr(AMRLevel* a_finer_level_ptr)
{
  m_finer_level_ptr = a_finer_level_ptr;
}

void AMRLevel::dt(Real a_dt)
{
  m_dt = a_dt;
}

Real AMRLevel::dt() const
{
  return(m_dt);
}

const ProblemDomain& AMRLevel::problemDomain() const
{
  return(m_problem_domain);
}

Vector<Box> AMRLevel::boxes() const
{
  return(m_level_grids);
}

int AMRLevel::refRatio() const
{
  return(m_ref_ratio);
}

void AMRLevel::time(Real a_time)
{
  m_time = a_time;
}

Real AMRLevel::time() const
{
  return m_time;
}

void AMRLevel::initialDtMultiplier(Real a_initial_dt_multiplier)
{
  m_initial_dt_multiplier = a_initial_dt_multiplier;
}

Real AMRLevel::initialDtMultiplier() const
{
  return m_initial_dt_multiplier;
}

// static
void AMRLevel::verbosity(int a_verbosity)
{
  s_verbosity = a_verbosity;
}

// static
int AMRLevel::verbosity()
{
  return(s_verbosity);
}

void AMRLevel::postRegrid(int a_base_level)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevel::postRegrid" << endl;
  }
}
