/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _CORNERCOPIER_H_
#define _CORNERCOPIER_H_

#include "DisjointBoxLayout.H"
#include "Pool.H"
#include "Vector.H"
#include "ProblemDomain.H"
#include "Copier.H"


/// An even stranger (than Copier) thing to copy from ghost cells to corner ghost cells.
/**
   This is a specialized Copier to fill corner ghost cells from ghost 
   cells of adjacent boxes.  Calling exchange with this copier will 
   fill _only_ the corner ghost cells.  All other exchange operations 
   should be performed by calling a separate exchange with a normal 
   Copier.   
*/

class CornerCopier: public Copier
{
public:

  ///null constructor, copy constructor and operator= can be compiler defined.
  CornerCopier()
  {}
  
  /// Basic defining constructor 
  /**  Note that this constructor doesn't take a ghost cell argument,
       so it creates a useless CornerCopier (no intersections are computed).
       Provided for compatibility with Copier base class.
  */
  CornerCopier(const DisjointBoxLayout& a_level, const BoxLayout& a_dest,
               bool a_exchange = false);
  
  /// this constructor contains support for periodic BC's
  /**  Note that this constructor doesn't take a ghost cell argument,
       so it creates a useless CornerCopier (no intersections are computed).
       Provided for compatibility with Copier base class.
  */
  CornerCopier(const DisjointBoxLayout& a_level, const BoxLayout& a_dest,
               const ProblemDomain& a_domain,
               bool  a_exchange  = false);
  
  /// Copier to copy into corner ghost cells of a_dest from ghost cells of  a_level
  /** Assumption is that a_dest and a_level have the same number of 
      ghost cells.
  */
  CornerCopier(const DisjointBoxLayout& a_level,
               const BoxLayout& a_dest,
               const IntVect& a_ghost, 
               bool  a_exchange = false);

  /// Copier to copy data into the valid and invalid regions of a_dest from a_level with support for periodic BCs
  /** Assumption is that a_dest and a_level have the same number of 
      ghost cells.
  */  
  CornerCopier(const DisjointBoxLayout& a_level,
               const BoxLayout& a_dest,
               const ProblemDomain& a_domain,
               const IntVect& a_ghost,
               bool  a_exchange = false);
  
  ///
  virtual ~CornerCopier();
  
  /// define to match constructor (also doesn't do anything)
  virtual void define(const DisjointBoxLayout& a_level, 
                      const BoxLayout& a_dest,
                      bool  a_exchange = false);

  /// define to match constructor (also doesn't do anything)
  virtual void define(const DisjointBoxLayout& a_level,
                      const BoxLayout& a_dest,
                      const ProblemDomain& a_domain,
                      bool  a_exchange = false);

  ///
  virtual void define(const DisjointBoxLayout& a_level,
                      const BoxLayout& a_dest,
                      const IntVect& a_ghost,
                      bool  a_exchange = false);

  /// contains support for periodic BCs
  virtual void define(const BoxLayout& a_level,
                      const BoxLayout& a_dest,
                      const ProblemDomain& a_domain,
                      const IntVect& a_ghost,
                      bool  a_exchange = false);

  /// 
  virtual void clear();


protected:
  
private:

};

std::ostream& operator<<(std::ostream& os, const CornerCopier& copier);


#endif
