/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _DATAINDEX_H_
#define _DATAINDEX_H_

#ifndef WRAPPER
class BoxLayout;

#include <cstdlib>

class DataIterator;
class LayoutIterator;

#endif /* WRAPPER */

///An index for LayoutIterator
class LayoutIndex
{
public:
  ///
  LayoutIndex();

  ~LayoutIndex()
  {}

  // default copy and assign should be fine.

  ///
  bool operator == (const LayoutIndex& a_rhs) const;

  bool eq(const LayoutIndex& a_rhs) const
  {
    return *this == a_rhs;
  }

  ///
  bool operator != (const LayoutIndex& a_rhs) const
  {
    return !(*this == a_rhs);
  }

  /// returns 'true' if this DataIndex has been null constructed.
  bool isNull() const;

  // not user function. breaks design encapsulation. method not intended for
  // users. The alternative is to make this class a friend of BoxLayoutData<T>,
  // which implicitly makes this class templated, which makes no bloody sense.
  // The behaviour of this method is undefined for users.  bvs
  int intCode() const;


private:
  friend class DataIterator;
  friend class LayoutIterator;
  friend class BoxLayout;
  friend class DataIndex;

  int        m_index;
  const int* m_layoutIntPtr;

  LayoutIndex(int        a_index,
              const int* a_layoutID)
    :
    m_index(a_index),
    m_layoutIntPtr(a_layoutID)
  {}
};

class DataIndex : public LayoutIndex
{
public:
  explicit DataIndex(const LayoutIndex& a_promotion)
    :
    LayoutIndex(a_promotion)
  {}

  DataIndex()
    :
    LayoutIndex()
  {}

  ~DataIndex()
  {}


private:
  friend class LayoutIterator;
  friend class DataIterator;
  friend class BoxLayout;
  friend class Copier;

  DataIndex(int        a_index,
            const int* a_layoutID)
    :
    LayoutIndex(a_index, a_layoutID)
  {}
};

#ifndef WRAPPER
inline LayoutIndex::LayoutIndex()
  :
  m_index(0),
  m_layoutIntPtr(NULL)
{}

inline bool LayoutIndex::operator == (const LayoutIndex& a_rhs) const
{
  return (m_index == a_rhs.m_index);
}

inline int LayoutIndex::intCode() const
{
  return m_index;
}

inline bool LayoutIndex::isNull() const
{
  return m_layoutIntPtr == NULL;
}
#endif

#endif
