/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _HDF5PORTABLE_H_
#define _HDF5PORTABLE_H_

#include "Vector.H"
extern "C"
{
#ifdef CH_USE_HDF5
#include <hdf5.h>
#else
  typedef long hid_t;
#endif
}

#include "Box.H"
#include "Interval.H"

template <class T> void dataTypes(Vector<hid_t>& a_types,
                                  const T&       a_dummySpecializationArg);

template <class T> void dataSize(const T&        a_item,
                                 Vector<int>&    a_sizes,
                                 const Box&      a_box,
                                 const Interval& a_comps);

template <class T> void write(const T&        a_item,
                              Vector<void*>&  a_allocatedBuffers,
                              const Box&      a_box,
                              const Interval& a_comps);

template <class T> void read(T&              a_item,
                             Vector<void*>&  a_allocatedBuffers,
                             const Box&      a_box,
                             const Interval& a_comps);

template <class T> const char* name(const T& a_dummySpecializationArg);

// These are the default template declarations for the HDF5 I/O
// template functions needed by T objects for binary IO.  The default
// implementations use the linearOut/linearIn/size functions of the
// T template classes to provide non-portable binary IO.
//
// A user can add their own template specializations for
// portable binary IO abilities.
//
// To use this header, a user includes this file first, then
// decalres their template specializations, then includes the CH_HDF5.H
// header file.
//
//   http://hdf.ncsa.uiuc.edu/HDF5/
//
// available datatypes in HDF5
//
//            Example
//                                                  Corresponding C Type
//            H5T_NATIVE_CHAR
//                                                  char
//            H5T_NATIVE_UCHAR
//                                                  unsigned char
//            H5T_NATIVE_SHORT
//                                                  short
//            H5T_NATIVE_USHORT
//                                                  unsigned short
//            H5T_NATIVE_INT
//                                                  int
//            H5T_NATIVE_UINT
//                                                  unsigned
//            H5T_NATIVE_LONG
//                                                  long
//            H5T_NATIVE_ULONG
//                                                  unsigned long
//            H5T_NATIVE_FLOAT
//                                                  float
//            H5T_NATIVE_DOUBLE
//                                                  double
//
// and some special ones defined for Chombo:
//
//            H5T_NATIVE_REAL                       Real
//
//            box_id                                Box
//
//            intvext_id                            IntVect

#endif
