/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _UGIO_H_
#define _UGIO_H_

#include <string>
using std::string;
using std::istream;

#include "LevelData.H"
#include "REAL.H"
#include "Vector.H"
#include "FArrayBox.H"
#include "IntVectSet.H"

#ifdef CH_USE_HDF5
/// Write single-level of data in HDF5 format
/**
   Writes a single level in HDF5 format.  Only available if the preprocessor
   macro HDF5 is defined at compilation.  This is blocking.

   \b Arguments:
   - \b a_filename: file to output to.
   - \b a_grid    : grids
   - \b a_data    : data
   - \b a_names   : names of variables.
   - \b a_domain  : domain of level.
*/
void WriteUGHDF5(const string&               a_filename,
                 const DisjointBoxLayout&    a_grids,
                 const LevelData<FArrayBox>& a_data,
                 const Box&                  a_domain);

/// Write a single level of data in HDF5 format.
/**
   Writes a single level in HDF5 format.  Only available if the preprocessor
   macro HDF5 is defined at compilation.  This is the hdf5 handle version and
   is not blocking.

   \b Arguments:
   - \b a_handle: handle to output to.
   - \b a_grid  : grids
   - \b a_data  : data
   - \b a_names : names of variables.
   - \b a_domain: domain of level.

*/
void WriteUGHDF5(HDF5Handle&                 a_handle,
                 const DisjointBoxLayout&    a_grids,
                 const LevelData<FArrayBox>& a_data,
                 const Box&                  a_domain);

/// Reads a level of data in HDF5 format.
/**
   Only available if the preprocessor macro HDF5 is defined at compilation.
   Returns 0 on success.  This is blocking.

   \b Arguments:
   - \b a_filename: file to input from.
   - \b a_grids   : grids
   - \b a_data    : data
   - \b a_domain  : domain

   \b Returns:
   - status code with values:
     -  0: success
     - -1: bogus number of levels
     - -2: bogus number of components
     - -3: error in readlevel
     - -4: file open failed
*/
int ReadUGHDF5(const string&         a_filename,
               DisjointBoxLayout&    a_grids,
               LevelData<FArrayBox>& a_data,
               Box&                  a_domain);

/// This is the handle version of ReadUGHDF5.
/**
   This is not blocking.
   Reads a  level of data in HDF5 format.  Only available if the
   preprocessor macro HDF5 is defined at compilation.  Returns 0 on success.

   \b Arguments:
   - \b a_handle: hdf5 handle to input from.
   - \b a_grids : grids
   - \b a_data  : data
   - \b a_domain: domain

   \b Returns:
   - status code with values:
      -  0: success
      - -1: bogus number of levels
      - -2: bogus number of components
      - -3: error in readlevel
      - -4: file open failed
*/
int
ReadUGHDF5(HDF5Handle&           a_handle,
           DisjointBoxLayout&    a_grids,
           LevelData<FArrayBox>& a_data,
           Box&                  a_domain);
#endif

#endif
