/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// BVS, July 10, 2003

#ifndef _BCFUNC_H_
#define _BCFUNC_H_

#include "FArrayBox.H"
#include "ProblemDomain.H"

typedef void(*BCFunc)(FArrayBox& a_state,
                                          const ProblemDomain& a_domain,
                                          Real a_dx,
                                          bool a_homogeneous);

extern "C" {

// given
//  pos [x,y,z] position on center of cell edge
//  int dir direction, x being 0
//  int side -1 for low, +1 = high,
//  fill in the a_values array

 typedef void(*BCValueFunc)(Real* pos,
                             int* dir,
                             int* side,
                             Real* a_values,
                             int*  a_numValues);

// given
//  pos [x,y,z] position on center of cell edge
//  int dir direction, x being 0
//  int side -1 for low, +1 = high,
//  int homogeneous.  false=0, true=1
//  Real* l1 (first valid point in FArrayBox) and l2,
//      fill in value of 'ghost'

  typedef void(*GenBCFunc)(Real* pos,
                           int* dir,
                           int* side,
                           int* homogeneous,
                           Real* l1,
                           Real* l2,
                           Real* ghost);
}

void GenBC(FArrayBox& a_state,
           const Box& a_valid,
           Real a_dx,
           bool a_homogeneous,
           GenBCFunc a_func,
           int a_dir,
           Side::LoHiSide a_side);

void NeumBC(FArrayBox& a_state,
            const Box& a_valid,
            Real a_dx,
            bool a_homogeneous,
            BCValueFunc a_value,
            int a_dir,
            Side::LoHiSide a_side);

void DiriBC(FArrayBox& a_state,
            const Box& valid,
            Real a_dx,
            bool a_homogeneous,
            BCValueFunc a_value,
            int a_dir,
            Side::LoHiSide a_side,
            int order = 1);

void DiriBC(FArrayBox& a_state,
            const Box& valid,
            Real a_dx,
            bool a_homogeneous,
            BCValueFunc a_value,
            int order = 1);

#endif
