/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// BVS, June 18, 2003

// We can assume that template class T has null construction.

#ifndef _NEWPOISSONOP_H_
#define _NEWPOISSONOP_H_

#include "AMRMultiGrid.H"
#include "REAL.H"
#include "Box.H"
#include "LevelDataOps.H"
#include "BCFunc.H"
#include "FArrayBox.H"
#include "NewCFIVS.H"

class NewPoissonOp : public MGLevelOp<FArrayBox>
{
public:
  NewPoissonOp(){ ; }

  virtual ~NewPoissonOp() { ; }
  /**
     \name LinearOp functions */
  /*@{*/

  void define(const RealVect& a_dx,
              const ProblemDomain& a_domain,
              BCFunc a_bc);

  virtual void createCoarsened(    FArrayBox& a_lhs,
                                   const FArrayBox& a_rhs,
                                   const int& a_refRat);

  virtual void residual(  FArrayBox& a_lhs,
                          const FArrayBox& a_phi,
                          const FArrayBox& a_rhs,
                          bool a_homogeneous = false);
  virtual void preCond(   FArrayBox& a_correction,
                          const FArrayBox& a_residual);
  virtual void applyOp(   FArrayBox& a_lhs,
                          const FArrayBox& a_phi,
                          bool a_homogeneous = false);
  virtual void create(    FArrayBox& a_lhs,
                          const FArrayBox& a_rhs);
  virtual void assign(    FArrayBox& a_lhs,
                          const FArrayBox& a_rhs) ;
  virtual Real dotProduct(const FArrayBox& a_1,
                          const FArrayBox& a_2) ;
  virtual void incr( FArrayBox& a_lhs,
                     const FArrayBox& a_x,
                     Real a_scale) ;
  virtual void axby( FArrayBox& a_lhs, const FArrayBox& a_x,
                     const FArrayBox& a_y,
                     Real a, Real b) ;
  virtual void scale(FArrayBox& a_lhs, const Real& a_scale) ;

  virtual Real norm(const FArrayBox& a_x, int a_ord);

  virtual void setToZero(FArrayBox& a_x);

  /*@}*/

  /**
     \name MGLevelOp functions */
  /*@{*/

  virtual void relax(FArrayBox& a_e,
                     const FArrayBox& a_residual,
                     int iterations);

  virtual void createCoarser(FArrayBox& a_coarse,
                             const FArrayBox& a_fine,
                             bool ghost);
  /**
     calculate restricted residual
     a_resCoarse[2h] = I[h->2h] (rhsFine[h] - L[h](phiFine[h])
  */
  virtual void restrictResidual(FArrayBox& a_resCoarse,
                                FArrayBox& a_phiFine,
                                const FArrayBox& a_rhsFine);

  /**
     correct the fine solution based on coarse correction
     a_phiThisLevel += I[2h->h](a_correctCoarse)
  */
  virtual void prolongIncrement(FArrayBox& a_phiThisLevel,
                                const FArrayBox& a_correctCoarse);

  /*@}*/


protected:
  RealVect                m_dx;
  RealVect                m_dxCrse;
  ProblemDomain           m_domain;
  BCFunc                  m_bc;

  void levelGSRB(FArrayBox& a_e,  const FArrayBox& a_residual);

};

class NewPoissonOpFactory : public MGLevelOpFactory<FArrayBox>
{
public:

  NewPoissonOpFactory();
  NewPoissonOpFactory(RealVect& a_dx, BCFunc a_bc);
  void define(const RealVect& m_dx,BCFunc m_bc);
  virtual ~NewPoissonOpFactory(){;}
  virtual NewPoissonOp* MGnewOp(const ProblemDomain& a_FineindexSpace,
                                   int   a_depth,
                                   bool  a_homoOnly = true);
  virtual void MGreclaim(NewPoissonOp* a_reclaim);

  RealVect                m_dx;
  BCFunc                  m_bc;

};

#endif
