/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#ifndef _INFINITEDOMAIN_F_H_
#define _INFINITEDOMAIN_F_H_

#include "FORT_PROTO.H"

extern "C"
{
#define FORT_DIFFPHI FORTRAN_NAME(DIFFPHI, diffphi)
  void FORT_DIFFPHI
  (
   double* phiex,
   double* phi,
   int* n,
   int* layer,
   double* h,
   int* vrbos,
   double* difnrm
   );


#define FORT_PRECONDRHS FORTRAN_NAME(PRECONDRHS, precondrhs)
  void FORT_PRECONDRHS
  (
   double* rhs,
   const int* l1, const int* l2, const int* l3,
   const int* u1, const int* u2, const int* u3
   );


#define FORT_INHOMODIRICHLET3D FORTRAN_NAME(INHOMODIRICHLET3D, inhomodirichlet3d)
  void FORT_INHOMODIRICHLET3D
  (
   double* f,
   const int* lx, const int* ly, const int* lz,
   const int* ux, const int* uy, const int* uz, 
   const double* dx, const double* dy, const double* dz,
   const int* op,
   int* ier
   );


#define FORT_POISSON3D FORTRAN_NAME(POISSON3D, poisson3d)
  void FORT_POISSON3D
  (
   double* f,
   int* n1,
   int* n2,
   int* n3,
   double* h,
   int* op
   );


#define FORT_INTERPFACE FORTRAN_NAME(INTERPFACE, interpface)
  void FORT_INTERPFACE
  (
   double* fine,
   double* coarse,
   double* intrpc,
   int* n1,
   int* n2,
   int* refToCoarse,
   int* n1Coarse,
   int* n2Coarse,
   int* layer
   );


#define FORT_GETPOLYINTERPC FORTRAN_NAME(GETPOLYINTERPC, getpolyinterpc)
  void FORT_GETPOLYINTERPC
  (
   double* intrpc,
   int* refToCoarse,
   int* layer
   );


#define FORT_GETBUFFERSIZE FORTRAN_NAME(GETBUFFERSIZE, getbuffersize)
  void FORT_GETBUFFERSIZE
  (
   int* domainLength,
   int* bufferSizeMin,
   int* bufferSize,
   int* patchSize,
   int* errCode
   );


#define FORT_GET0BUFFERSIZE FORTRAN_NAME(GET0BUFFERSIZE, get0buffersize)
  void FORT_GET0BUFFERSIZE
  (
   int* domainLength,
   int* bufferSizeMin,
   int* bufferSize,
   int* patchSize,
   int* errCode
   );


#define FORT_GET0BUFFERSIZE3D FORTRAN_NAME(GET0BUFFERSIZE3D, get0buffersize3d)
  void FORT_GET0BUFFERSIZE3D
  (
   int* domainLength,
   int* bufferSizeMin,
   int* bufferSize,
   int* patchSize,
   int* errCode
   );


#define FORT_GETBUFFERSIZE3D FORTRAN_NAME(GETBUFFERSIZE3D, getbuffersize3d)
  void FORT_GETBUFFERSIZE3D
  (
   int* domainLength,
   int* bufferSizeMin,
   int* bufferSize,
   int* patchSize,
   int* errCode
   );


#define FORT_GETCBUFFERSIZE FORTRAN_NAME(GETCBUFFERSIZE, getcbuffersize)
  void FORT_GETCBUFFERSIZE
  (
   int* domainLength,
   int* bufferSizeMin,
   int* bufferSize,
   int* patchSize,
   int* errCode
   );


#define FORT_COEFFS1SIDEDDERIV FORTRAN_NAME(COEFFS1SIDEDDERIV, coeffs1sidedderiv)
  void FORT_COEFFS1SIDEDDERIV
  (
   double* coeff,
   int* d,
   double* h
   );


#define FORT_COEFFSDERIV FORTRAN_NAME(COEFFSDERIV, coeffsderiv)
  void FORT_COEFFSDERIV
  (
   double* coeff,
   int* d,
   double* h,
   double* start,
   double* evalpt
   );


#define FORT_DERIV1OVERR FORTRAN_NAME(DERIV1OVERR, deriv1overr)
  void FORT_DERIV1OVERR
  (
   int* cpind,
   double* cfac,
   int* cpow,
   int* nterms,
   int* deg
   );


#define FORT_GETRECTMATRIX FORTRAN_NAME(GETRECTMATRIX, getrectmatrix)
  void FORT_GETRECTMATRIX
  (
   double* srcdst,
   const int* iev,
   const int* icp,
   const int* nsrc,
   const int* ndst,
   const int* refc,
   const int* ptchsz,
   const int* npchs1,
   const int* npchs2,
   const int* ndst1c,
   const int* ndst2c,
   const int* deg,
   const int* layer,
   const double* h,
   const int* cpind,
   const double* cfac,
   const int* cpow,
   const int* nterms,
   const int* vrbos,
   const int* ier
   );


#define FORT_GETORIRECTMATRIX FORTRAN_NAME(GETORIRECTMATRIX, getorirectmatrix)
  void FORT_GETORIRECTMATRIX
  (
   const int* ori,
   double* srcdst,
   const int* iev,
   const int* icp,
   const int* nsrc,
   const int* ndst,
   const int* refc,
   const int* ptchsz,
   const int* npchs1,
   const int* npchs2,
   const int* ndst1c,
   const int* ndst2c,
   const int* deg,
   const int* layer,
   const double* h,
   const int* cpind,
   const double* cfac,
   const int* cpow,
   const int* nterms,
   const int* vrbos,
   const int* ier
   );


#define FORT_GETDIRECTRECTMATRIX FORTRAN_NAME(GETDIRECTRECTMATRIX, getdirectrectmatrix)
  void FORT_GETDIRECTRECTMATRIX
  (
   double* srcdst,
   const double* wtsint,
   const int* iev,
   const int* icp,
   const int* nsrc,
   const int* ndst1c,
   const int* ndst2c,
   const int* layer,
   const double* hsrc,
   const double* hdst,
   const double* cpbase,
   const double* evbase,
   const int* vrbos,
   const int* ier
   );


#define FORT_GETALLCOARSEMATRIX FORTRAN_NAME(GETALLCOARSEMATRIX, getallcoarsematrix)
  void FORT_GETALLCOARSEMATRIX
  (
   double* sdcrse,
   int* ncpts,
   int* n1,
   int* s2,
   int* ptchsz,
   int* npchs,
   int* deg,
   double* h,
   int* refc,
   int* cadd,
   int* cpind,
   double* cfac,
   int* cpow,
   int* nterms,
   int* vrbos,
   int* ier
   );


#define FORT_GETRECTCOARSEMATRIX FORTRAN_NAME(GETRECTCOARSEMATRIX, getrectcoarsematrix)
  void FORT_GETRECTCOARSEMATRIX
  (
   double* sdcrse,
   int* icp,
   int* nsrc,
   int* ndst,
   int* refc,
   int* ptchsz,
   int* npchs1,
   int* npchs2,
   int* deg,
   double* h,
   int* cadd,
   int* cpind,
   double* cfac,
   int* cpow,
   int* nterms,
   int* ncpts,
   int* vrbos,
   int* ier
   );


#define FORT_ADDEVAL1FACECOARSE FORTRAN_NAME(ADDEVAL1FACECOARSE, addeval1facecoarse)
  void FORT_ADDEVAL1FACECOARSE
  (
   double* ev,
   const double* cp,
   const int* npchs1,
   const int* npchs2,
   const int* ndst1c,
   const int* ndst2c,
   const int* nc,
   const int* layer,
   const double* srcdst,
   const int* vrbos,
   int* ier
   );


#define FORT_ADDEVALSYM1FACECOARSE FORTRAN_NAME(ADDEVALSYM1FACECOARSE, addevalsym1facecoarse)
  void FORT_ADDEVALSYM1FACECOARSE
  (
   double* ev,
   const double* cp,
   const int* dflip,
   const int* sflip,
   const int* npchs1,
   const int* npchs2,
   const int* ndst1c,
   const int* ndst2c,
   const int* deg,
   const int* nc,
   const int* layer,
   const double* srcdst,
   const int* vrbos,
   int* ier
   );


#define FORT_ADDEVAL1DIRECTFACECOARSE FORTRAN_NAME(ADDEVAL1DIRECTFACECOARSE, addeval1directfacecoarse)
  void FORT_ADDEVAL1DIRECTFACECOARSE
  (
   double* ev,
   const double* cp,
   const int* nsrc,
   const int* ndst1c,
   const int* ndst2c,
   const int* layer,
   const double* srcdst,
   const int* vrbos,
   int* ier
   );


#define FORT_EVALOUTERCOARSE FORTRAN_NAME(EVALOUTERCOARSE, evaloutercoarse)
  void FORT_EVALOUTERCOARSE
  (
   double* ev,
   const double* cp,
   const int* ncpts,
   const int* npchs,
   const int* n2c,
   const int* cadd,
   const int* nc,
   const int* ori,
   const double* sdcrse,
   const int* vrbos,
   int* ier
   );


#define FORT_EVALRECTOUTERCOARSE FORTRAN_NAME(EVALRECTOUTERCOARSE, evalrectoutercoarse)
  void FORT_EVALRECTOUTERCOARSE
  (
   double* ev,
   const double* cp,
   const int* ncpts,
   const int* npchs1,
   const int* npchs2,
   const int* ndstc,
   const int* cadd,
   const int* nc,
   const double* sdcrse,
   const int* vrbos,
   int* ier
   );


#define FORT_EVALSYMRECTOUTERCOARSE FORTRAN_NAME(EVALSYMRECTOUTERCOARSE, evalsymrectoutercoarse)
  void FORT_EVALSYMRECTOUTERCOARSE
  (
   double* ev,
   const double* cp,
   const int* dflip,
   const int* ncpts,
   const int* npchs1,
   const int* npchs2,
   const int* ndstc,
   const int* cadd,
   const int* nc,
   const double* sdcrse,
   const int* vrbos,
   int* ier
   );


#define FORT_EVALDIRECTRECTOUTERCOARSE FORTRAN_NAME(EVALDIRECTRECTOUTERCOARSE, evaldirectrectoutercoarse)
  void FORT_EVALDIRECTRECTOUTERCOARSE
  (
   double* ev,
   const double* cp,
   const int* ncpts,
   const int* ndstc,
   const int* cadd,
   const int* nsrc,
   const double* sdcrse,
   const int* vrbos,
   int* ier
   );

#define FORT_GETRECTDIRECTCOARSEMATRIX FORTRAN_NAME(GETRECTDIRECTCOARSEMATRIX, getrectdirectcoarsematrix)
  void FORT_GETRECTDIRECTCOARSEMATRIX
  (
   double* sdcrse,
   const double* wtsint,
   const int* ncpts,
   const int* ndstc,
   const int* cadd,
   const int* nsrc,
   const double* hsrc,
   const double* hdst,
   const double* cpbase,
   const double* evbase,
   const int* vrbos,
   const int* ier
   );

}

#endif
