/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

#include <cstdio>
//[NOTE: the order of these is important because memtrack.H and Arena.H conflict. <dbs>]
#include "parstream.H"
#include "MayDay.H"
#include "CH_HDF5.H"
#include "memtrack.H"
#include "memusage.H"

// note:  i copied the following #include crud from elsewhere.
// assumed it was still relevant (ndk)
#include <sys/time.h>
#ifdef __SOLARIS2__
/* This is a problem.  Version 2.5 of Solaris (at least) does not have
 * rusage.h, preferring to include the definition of struct rusage in
 * resource.h.  Version 2.3 of Solaris (at least) has resource.h but
 * it does not include struct rusage, which is rather defined in rusage.h
 *
 * The _right_ fix
 */
#  ifdef USE_RUSAGE_H
#    include <sys/rusage.h>
#  else
#    include <sys/resource.h>
#  endif
#  include <sys/procfs.h>
#else
#  include <sys/resource.h>
#endif

#ifdef __hpux
#include <sys/syscall.h>
#define getrusage(a,b) syscall(SYS_getrusage,a,b)
#endif

void print_memory_line(void) {

#ifdef CH_USE_MEMORY_TRACKING
  Real memtrackCurrentMemory;
  Real memtrackPeakMemory;
  memtrackStamp(memtrackCurrentMemory, memtrackPeakMemory);

  char temp[180];
  sprintf(temp, " Mem Usage: OS=%8.3f  MT_peak=%8.3f  MT_current=%8.3f (MB)\n",
          get_memory_usage_from_OS(), memtrackPeakMemory, memtrackCurrentMemory);
  pout() << temp;
#else
  char temp[180];
  sprintf(temp, " Mem Usage: OS=%8.3f(MB)  MT is off\n",
          get_memory_usage_from_OS());
  pout() << temp;
#endif

}

void print_memory_line(char *s) {

#ifdef CH_USE_MEMORY_TRACKING
  Real memtrackCurrentMemory;
  Real memtrackPeakMemory;
  memtrackStamp(memtrackCurrentMemory, memtrackPeakMemory);

  char temp[180];
  sprintf(temp, "%26s|Mem Usage: OS=%8.3f  MT_peak=%8.3f  MT_current=%8.3f (MB)\n",
          s, get_memory_usage_from_OS(), memtrackPeakMemory, memtrackCurrentMemory);
  pout() << temp;
#else
  char temp[180];
  sprintf(temp, "%26s|Mem Usage: OS=%8.3f (MB)  MT is off\n",
          s, get_memory_usage_from_OS());
  pout() << temp;
#endif

}

// returns memory used in MB from an OS system call
// still unsure if i should be dividing by 1000 or 1024
// to get MB...
Real get_memory_usage_from_OS(void) {

  Real memusage = 0.0;

  // if linux, use a different method.
#ifdef CH_Linux
  FILE* f=fopen("/proc/self/statm","r");
  int size, resident, shared;
  if(fscanf(f, "%d %d %d", &size, &resident, &shared)==3) {
    //printf("%10.2f MB total size, %10.2f MB resident, %10.2f MB shared\n",
    //     (size*4)/1000.0, (resident*4)/1000.0, (shared*4)/1000.0);
    memusage = (size*4)/1000.0;  // ??
  }
  fclose(f);
#else
  static struct rusage rus;
  getrusage(RUSAGE_SELF, &rus);

  // ru_maxrss is a long int and is the "Maximum resident set size (in kilobytes)"

  //printf(" ru_maxrss= %d  %10.2fMB\n",
  // rus.ru_maxrss, rus.ru_maxrss/1000.0);
  //printf(" ru_ixrss=%d ru_idrss=%d ru_isrs=%d ru_maxrss=%d\n",
  //     rus.ru_ixrss, rus.ru_idrss, rus.ru_isrss, rus.ru_maxrss);
  memusage = rus.ru_maxrss/1000.0;
#endif

  return memusage;

}

#if defined(CH_MPI)

#include <iostream>
using std::endl;
#include "SPMD.H"

// input is the "end" memory (like the peak) and the output
// is the avg, min, and max memory across all processors
void gather_memory_from_procs(Real end_memory,
                              Real &avg_memory,
                              Real &min_memory,
                              Real &max_memory) {

  // Depending on the mechanism used in get_memory_usage() to report
  // memory use, the variable end_memory will either be the peak
  // memory used at any point in the code execution (as getrusage()
  // works on OSF machine such as halem) or an estimate of the amount
  // of memory currently being used (which will certainly not be the
  // peak).  If end_memory is the peak memory used so far, then it is
  // useful in parallel jobs to gather these peak memory usage values
  // from all processors and obtain the minimum, maximum, and average
  // values.  This is what the following code does. (ndk)
  int result;
  Real sum_memory;
  result = MPI_Reduce(&end_memory, &sum_memory, 1, MPI_CH_REAL,
                      MPI_SUM, 0, Chombo_MPI::comm);
  if (result != MPI_SUCCESS)
    {
      MayDay::Error("Sorry, but I had a communication error in getting memorys");
    }

  result = MPI_Reduce(&end_memory, &min_memory, 1, MPI_CH_REAL,
                      MPI_MIN, 0, Chombo_MPI::comm);
  if (result != MPI_SUCCESS)
    {
      MayDay::Error("Sorry, but I had a communication error in getting memorys");
    }

  result = MPI_Reduce(&end_memory, &max_memory, 1, MPI_CH_REAL,
                      MPI_MAX, 0, Chombo_MPI::comm);
  if (result != MPI_SUCCESS)
    {
      MayDay::Error("Sorry, but I had a communication error in getting memorys");
    }

  int rank, number_procs;
  MPI_Comm_rank(Chombo_MPI::comm, &rank);
  MPI_Comm_size(Chombo_MPI::comm, &number_procs);

  // results only need to be on rank0 (and are)
  if (rank==0)
    {
      Real avg_memory = sum_memory/(Real)number_procs;

      pout() << "Gather end memory from procs:  avg: " << avg_memory
             << "  min: " << min_memory
             << "  max: " << max_memory << " (MB)" << endl;

    }
}
#endif
