/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  V \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
// CHOMBO Copyright (c) 2000-2004, The Regents of the University of
// California, through Lawrence Berkeley National Laboratory (subject to
// receipt of any required approvals from U.S. Dept. of Energy).  All
// rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// (1) Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// (2) Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
// (3) Neither the name of Lawrence Berkeley National Laboratory, U.S.
// Dept. of Energy nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
// OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// You are under no obligation whatsoever to provide any bug fixes,
// patches, or upgrades to the features, functionality or performance of
// the source code ("Enhancements") to anyone; however, if you choose to
// make your Enhancements available either publicly, or directly to
// Lawrence Berkeley National Laboratory, without imposing a separate
// written license agreement for such Enhancements, then you hereby grant
// the following license: a non-exclusive, royalty-free perpetual license
// to install, use, modify, prepare derivative works, incorporate into
// other computer software, distribute, and sublicense such Enhancements or
// derivative works thereof, in binary and source code form.
//
// TRADEMARKS. Product and company names mentioned herein may be the
// trademarks of their respective owners.  Any rights not expressly granted
// herein are reserved.
//

// testParmParse.cpp

// tests ParmParse.  using the input file inputs.testParmParse you
// should get the output in 2d:
//
//   testing query functions
//   foo1 = 1
//   foo2 = 2.5
//   pp.countname(foo2_name_stds) = 2
//   foo3 = 3
//   foo4stdv[0] = 4
//   foo4stdv[1] = 5
//   foo5_stds = 'hello'
//   foo6stdvs[0] = 'man'
//   foo6stdvs[1] = 'plan'
//   foo7stdv[0] = 7
//   foo7stdv[1] = 8
//
//   testing get functions
//   foo1 = 1
//   foo2 = 2.5
//   pp.countname(foo2_name_stds) = 2
//   foo3 = 3
//   foo4stdv[0] = 4
//   foo4stdv[1] = 5
//   foo5_stds = 'hello'
//   foo6stdvs[0] = 'man'
//   foo6stdvs[1] = 'plan'
//   foo7stdv[0] = 7
//   foo7stdv[1] = 8
//
// and in 3d you should get:
//
//   testing query functions
//   foo1 = 1
//   foo2 = 2.5
//   pp.countname(foo2_name_stds) = 2
//   foo3 = 3
//   foo4stdv[0] = 4
//   foo4stdv[1] = 5
//   foo4stdv[2] = 6
//   foo5_stds = 'hello'
//   foo6stdvs[0] = 'man'
//   foo6stdvs[1] = 'plan'
//   foo6stdvs[2] = 'canal'
//   foo7stdv[0] = 7
//   foo7stdv[1] = 8
//   foo7stdv[2] = 9
//
//   testing get functions
//   foo1 = 1
//   foo2 = 2.5
//   pp.countname(foo2_name_stds) = 2
//   foo3 = 3
//   foo4stdv[0] = 4
//   foo4stdv[1] = 5
//   foo4stdv[2] = 6
//   foo5_stds = 'hello'
//   foo6stdvs[0] = 'man'
//   foo6stdvs[1] = 'plan'
//   foo6stdvs[2] = 'canal'
//   foo7stdv[0] = 7
//   foo7stdv[1] = 8
//   foo7stdv[2] = 9
//

#include <iostream>
#include <iomanip>
#include <string>

#include "SPACE.H"
#include "REAL.H"
#include "ParmParse.H"
#ifdef CH_MPI
#include <mpi.h>
#endif
/// Prototypes:

using std::cout;
using std::endl;
using std::boolalpha;

void
parseTestOptions(int argc ,char* argv[]) ;

int
testParmParse(int argc, char* argv[]);

void
testQuery();

void
testGet();

/// Global variables for handling output:

static const char *pgmname = "testParmParse" ;
static const char *indent = "   " ,*indent2 = "      " ;
static bool verbose = true ;

/// Code:

int
main(int argc, char* argv[])
{
#ifdef CH_MPI
  MPI_Init(&argc, &argv);
#endif
  parseTestOptions( argc ,argv ) ;
  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  int ret = testParmParse(argc, argv);
#ifdef CH_MPI
  MPI_Finalize();
#endif
  return ret;
}

int
testParmParse(int argc, char* argv[])
{
  // use the first argument that isn't an option as the input filename
  // and take it out of the argument list
  char* in_file = NULL ;
  for( int i=1 ; i<argc ; ++i )
    {
      if( strlen(argv[i]) > 0 && argv[i][0] != '-' ) {
        in_file = argv[i] ;
        argv[i] = "" ;
        break ;
      }
    }

  if( in_file == NULL )
    {
      in_file = "inputs.testParmParse";
    }

  ParmParse pp(argc-1, argv+1, NULL, in_file);

  if(verbose)
    cout << indent2 << pgmname
         << ": testing query functions" << endl;
  testQuery();
  if(verbose)
    cout << indent2 << pgmname
         << ": testing get functions" << endl;
  testGet();
  cout << indent << pgmname
       << ": ParmParse tests passed." << endl ;

  return 0 ;
}

void
testQuery()
{
    ParmParse pp;

    int foo1 = 0;
    pp.query("foo1", foo1);
    if(verbose) cout << indent2 << pgmname << ": query int foo1 = " << foo1 << endl;

    Real foo2 = 0;
    pp.query("foo2", foo2);
    if(verbose) cout << indent2 << pgmname << ": query Real foo2 = " << foo2 << endl;

    std::string foo2_name_stds("foo2");
    int num_foo2_stds = pp.countname(foo2_name_stds);
    if(verbose) cout << indent2 << pgmname << ": std::string pp.countname(foo2) = " << num_foo2_stds << endl;

    Real foo3 = 0;
    pp.query("foo3", foo3);
    if(verbose) cout << indent2 << pgmname << ": query Real foo3 = " << foo3 << endl;

    std::vector<Real> foo4stdv(SpaceDim);
    pp.queryarr("foo4", foo4stdv, 0, SpaceDim);
    if(verbose)
      {
        cout << indent2 << pgmname << ": queryarr std::vector<Real> foo4 = " << endl ;
        for (int d = 0; d < SpaceDim; ++d)
          {
            cout << indent2 << "[" << d << "] = " << foo4stdv[d] << endl;
          }
      }

    std::string foo5_stds;
    pp.query("foo5", foo5_stds);
    if(verbose) cout << indent2 << pgmname << ": query std::string foo5 = '" << foo5_stds << "'" << endl;

    std::vector<std::string> foo6stdvs;
    pp.queryarr("foo6", foo6stdvs, 0, SpaceDim);
    if(verbose)
      {
        cout << indent2 << pgmname << ": queryarr std::vector<std::string> foo6 =" << endl ;
        for (int d = 0; d < SpaceDim; ++d)
          {
            cout << indent2 << "[" << d << "] = '" << foo6stdvs[d] << "'" << endl;
          }
      }

    std::vector<int> foo7stdv(SpaceDim);
    pp.queryarr("foo7", foo7stdv, 0, SpaceDim);
    if(verbose)
      {
        cout << indent2 << pgmname << ": queryarr std::vector<int> foo7 =" << endl ;
        for (int d = 0; d < SpaceDim; ++d)
          {
            cout << indent2 << "[" << d << "] = " << foo7stdv[d] << endl;
          }
      }

    cout << boolalpha ; // output symbolic representation of boolean (true,false) instead of (1,0)
    bool foobool1=false,foobool2=false,foobool3=true,foobool4=true,foobool5=false,foobool6=true;
    pp.query("foobool1",foobool1);
    pp.query("foobool2",foobool2);
    pp.query("foobool3",foobool3);
    pp.query("foobool4",foobool4);
    pp.query("foobool5",foobool5);
    pp.query("foobool6",foobool6);
    if(verbose)
      {
        cout << indent2 << pgmname << ": query foobool1: input=true,  expected result true,  result = " << foobool1 << endl ;
        cout << indent2 << pgmname << ": query foobool2: input=1,     expected result true,  result = " << foobool2 << endl ;
        cout << indent2 << pgmname << ": query foobool3: input=false, expected result false, result = " << foobool3 << endl ;
        cout << indent2 << pgmname << ": query foobool4: input=0,     expected result false, result = " << foobool4 << endl ;
        cout << indent2 << pgmname << ": query foobool5: input=TRUE,  expected result true,  result = " << foobool5 << endl ;
        cout << indent2 << pgmname << ": query foobool6: input=FALSE, expected result false, result = " << foobool6 << endl ;
      }
    if( !( foobool1 && foobool2 && ! foobool3 && ! foobool4 && foobool5 && ! foobool6 ) ){
      MayDay::Error( "bool query test failed" ,1 );
    }
}

void
testGet()
{
    ParmParse pp;

    int foo1 = 0;
    pp.get("foo1", foo1);
    if(verbose) cout << indent2 << pgmname << ": get int foo1 = " << foo1 << endl;

    Real foo2 = 0;
    pp.get("foo2", foo2);
    if(verbose) cout << indent2 << pgmname << ": get Real foo2 = " << foo2 << endl;

    std::string foo2_name_stds("foo2");
    int num_foo2_stds = pp.countname(foo2_name_stds);
    if(verbose) cout << indent2 << pgmname << ": std::string pp.countname(foo2) = " << num_foo2_stds << endl;

    Real foo3 = 0;
    pp.get("foo3", foo3);
    if(verbose) cout << indent2 << pgmname << ": get Real foo3 = " << foo3 << endl;

    std::vector<Real> foo4stdv(SpaceDim);
    pp.getarr("foo4", foo4stdv, 0, SpaceDim);
    if(verbose)
      {
        cout << indent2 << pgmname << ": queryarr std::vector<Real> foo4 = " << endl ;
        for (int d = 0; d < SpaceDim; ++d)
          {
            cout << indent2 << "[" << d << "] = " << foo4stdv[d] << endl;
          }
      }

    std::string foo5_stds;
    pp.get("foo5", foo5_stds);
    if(verbose) cout << indent2 << pgmname << ": get std::string foo5 = '" << foo5_stds << "'" << endl;

    std::vector<std::string> foo6stdvs;
    pp.getarr("foo6", foo6stdvs, 0, SpaceDim);
    if(verbose)
      {
        cout << indent2 << pgmname << ": getarr std::vector<std::string> foo6 =" << endl ;
        for (int d = 0; d < SpaceDim; ++d)
          {
            cout << indent2 << "[" << d << "] = '" << foo6stdvs[d] << "'" << endl;
          }
      }

    std::vector<int> foo7stdv(SpaceDim);
    pp.getarr("foo7", foo7stdv, 0, SpaceDim);
    if(verbose)
      {
        cout << indent2 << pgmname << ": getarr std::vector<int> foo7 =" << endl ;
        for (int d = 0; d < SpaceDim; ++d)
          {
            cout << indent2 << "[" << d << "] = " << foo7stdv[d] << endl;
          }
      }
}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
